"""
Core functionality for the graphlib package.
"""
from typing import Optional

from relationalai.early_access.builder import Concept, Relationship
from relationalai.early_access.builder import Integer#, Float
from relationalai.early_access.builder import where, define, count

class Graph():
    def __init__(self,
            *,
            directed: bool,
            weighted: bool,
            aggregator: Optional[str] = None,
        ):
        assert isinstance(directed, bool), "The `directed` argument must be a boolean."
        assert isinstance(weighted, bool), "The `weighted` argument must be a boolean."
        self.directed = directed
        self.weighted = weighted

        assert isinstance(aggregator, type(None)), "Weight aggregation not yet supported."
        # TODO: In the hopefully not-too-distant future, this argument will
        #   allow the user to specify whether and how to aggregate weights
        #   for multi-edges that exist at the user interface (Edge) level
        #   to construct the internal edge/weight list representation.
        #   The `str` type is just a placeholder; it should be something else.

        # Introduce Node and Edge concepts.
        Node = Concept("Node")
        Edge = Concept("Edge")
        Edge.src = Relationship("{edge:Edge} has source {src:Node}")
        Edge.dst = Relationship("{edge:Edge} has destination {dst:Node}")
        Edge.weight = Relationship("{edge:Edge} has weight {weight:Float}")
        self.Node = Node
        self.Edge = Edge

        # TODO: Require that each Edge has an Edge.src.
        # TODO: Require that each Edge has an Edge.dst.
        # TODO: If weighted, require that each Edge has an Edge.weight.
        # TODO: If not weighted, require that each Edge does not have an Edge.weight.

        # TODO: Suppose that type checking should in future restrict `src` and
        #   `dst` to be `Node`s, but at the moment we may need a require for that.
        # TODO: Suppose that type checking should in future restrict `weight` to be
        #   `Float`s, but at the moment we may need a require for that.

        # TODO: Transform Node and Edge into underlying edge-/weight-list representation.
        # NOTE: Operate under the assumption that `Node` contains all
        #   possible nodes, i.e. we can use the `Node` Concept directly as
        #   the node list. Has the additional benefit of allowing relationships
        #   (for which it makes sense) to be properties of `Node` rather than standalone.
        self._define_edge_relationships()
 
        self._define_num_nodes_relationship()
        self._define_num_edges_relationship()

        self._define_neighbor_relationships()
        self._define_count_neighbor_relationships()
        self._define_common_neighbor_relationship()
        self._define_count_common_neighbor_relationship()

        self._define_degree_relationships()


    def _define_edge_relationships(self):
        """
        Define the self._edge and self._weight relationships,
        consuming the Edge concept's `src`, `dst`, and `weight` relationships.
        """
        self._edge = Relationship("{src:Node} has edge to {dst:Node}")
        self._weight = Relationship("{src:Node} has edge to {dst:Node} with weight {weight:Float}")

        Edge = self.Edge
        if self.directed and self.weighted:
            define(
                self._weight(Edge.src, Edge.dst, Edge.weight),
                self._edge(Edge.src, Edge.dst)
            )
        elif self.directed and not self.weighted:
            define(
                self._weight(Edge.src, Edge.dst, 1.0),
                self._edge(Edge.src, Edge.dst)
            )
        elif not self.directed and self.weighted:
            define(
                self._weight(Edge.src, Edge.dst, Edge.weight),
                self._weight(Edge.dst, Edge.src, Edge.weight),
                self._edge(Edge.src, Edge.dst),
                self._edge(Edge.dst, Edge.src)
            )
        elif not self.directed and not self.weighted:
            define(
                self._weight(Edge.src, Edge.dst, 1.0),
                self._weight(Edge.dst, Edge.src, 1.0),
                self._edge(Edge.src, Edge.dst),
                self._edge(Edge.dst, Edge.src)
            )

    def _define_num_nodes_relationship(self):
        """Define the self._num_nodes relationship."""
        self._num_nodes = Relationship("The graph has {num_nodes:Integer} nodes")
        define(self._num_nodes(count(self.Node) | 0))

    def _define_num_edges_relationship(self):
        """Define the self._num_edges relationship."""
        self._num_edges = Relationship("The graph has {num_edges:Integer} edges")

        src, dst = self.Node.ref(), self.Node.ref()

        if self.directed:
            define(self._num_edges(count(src, dst, self._edge(src, dst)) | 0))
        elif not self.directed:
            define(self._num_edges(count(src, dst, self._edge(src, dst), src <= dst) | 0))
            # TODO: Generates an UnresolvedOverload warning from the typer.
            #   Should be sorted out by improvements in the typer (to allow
            #   comparisons between instances of concepts).


    def _define_neighbor_relationships(self):
        """Define the self.[in,out]neighbor relationships."""
        self._neighbor = Relationship("{src:Node} has neighbor {dst:Node}")
        self._inneighbor = Relationship("{dst:Node} has inneighbor {src:Node}")
        self._outneighbor = Relationship("{src:Node} has outneighbor {dst:Node}")

        src, dst = self.Node.ref(), self.Node.ref()
        where(self._edge(src, dst)).define(self._neighbor(src, dst), self._neighbor(dst, src))
        where(self._edge(dst, src)).define(self._inneighbor(src, dst))
        where(self._edge(src, dst)).define(self._outneighbor(src, dst))
        # Note that these definitions happen to work for both
        # directed and undirected graphs due to `edge` containing
        # each edge's symmetric partner in the undirected case.

    def _define_count_neighbor_relationships(self):
        """
        Define the self._count_[in,out]neighbor relationships.
        Note that these relationships differ from corresponding
        [in,out]degree relationships in that they yield empty
        rather than zero absent [in,out]neighbors.
        Primarily for internal consumption.
        """
        self._count_neighbor = Relationship("{src:Node} has neighbor count {count:Integer}")
        self._count_inneighbor = Relationship("{dst:Node} has inneighbor count {count:Integer}")
        self._count_outneighbor = Relationship("{src:Node} has outneighbor count {count:Integer}")

        src, dst = self.Node.ref(), self.Node.ref()
        where(self._neighbor(src, dst)).define(self._count_neighbor(src, count(dst).per(src)))
        where(self._inneighbor(dst, src)).define(self._count_inneighbor(dst, count(src).per(dst)))
        where(self._outneighbor(src, dst)).define(self._count_outneighbor(src, count(dst).per(src)))


    def _define_common_neighbor_relationship(self):
        """Define the self._common_neighbor relationship."""
        self._common_neighbor = Relationship("{node_a:Node} and {node_b:Node} have common neighbor {node_c:Node}")

        node_a, node_b, node_c = self.Node.ref(), self.Node.ref(), self.Node.ref()
        where(self._neighbor(node_a, node_c), self._neighbor(node_b, node_c)).define(self._common_neighbor(node_a, node_b, node_c))

    def _define_count_common_neighbor_relationship(self):
        """Define the self._count_common_neighbor relationship."""
        self._count_common_neighbor = Relationship("{node_a:Node} and {node_b:Node} have common neighbor count {count:Integer}")

        node_a, node_b, node_c = self.Node.ref(), self.Node.ref(), self.Node.ref()
        where(self._common_neighbor(node_a, node_b, node_c)).define(self._count_common_neighbor(node_a, node_b, count(node_c).per(node_a, node_b)))


    def _define_degree_relationships(self):
        """Define the self._[in,out]degree relationships."""
        self._degree = Relationship("{node:Node} has degree {count:Integer}")
        self._indegree = Relationship("{node:Node} has indegree {count:Integer}")
        self._outdegree = Relationship("{node:Node} has outdegree {count:Integer}")

        incount, outcount = Integer.ref(), Integer.ref()

        where(
            _indegree := where(self._count_inneighbor(self.Node, incount)).select(incount),
        ).define(self._indegree(self.Node, _indegree | 0))

        where(
            _outdegree := where(self._count_outneighbor(self.Node, outcount)).select(outcount),
        ).define(self._outdegree(self.Node, _outdegree | 0))

        if self.directed:
            where(
                _indegree := where(self._indegree(self.Node, incount)).select(incount),
                _outdegree := where(self._outdegree(self.Node, outcount)).select(outcount),
            ).define(self._degree(self.Node, (_indegree | 0) + (_outdegree | 0)))
        elif not self.directed:
            neighcount = Integer.ref()
            where(
                _degree := where(self._count_neighbor(self.Node, neighcount)).select(neighcount),
            ).define(self._degree(self.Node, _degree | 0))


    # Public accessor methods for private relationships.
    def num_nodes(self): return self._num_nodes
    def num_edges(self): return self._num_edges

    def neighbor(self): return self._neighbor
    def inneighbor(self): return self._inneighbor
    def outneighbor(self): return self._outneighbor

    def count_neighbor(self): return self._count_neighbor
    def count_inneighbor(self): return self._count_inneighbor
    def count_outneighbor(self): return self._count_outneighbor

    def common_neighbor(self): return self._common_neighbor
    def count_common_neighbor(self): return self._count_common_neighbor

    def degree(self): return self._degree
    def indegree(self): return self._indegree
    def outdegree(self): return self._outdegree



from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass
import datetime
from typing import Optional, Union
from relationalai import debugging
from relationalai.early_access.metamodel import builtins, ir, types, visitor, compiler, factory as f
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
import rich

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def to_relation(node:ir.Lookup|ir.Update|ir.Aggregate) -> ir.Relation:
    if isinstance(node, ir.Aggregate):
        return node.aggregation
    return node.relation

def to_name(type:ir.Type) -> str:
    if isinstance(type, ir.ScalarType):
        return type.name
    elif isinstance(type, ir.UnionType):
        return '|'.join([to_name(t) for t in type.types])
    else:
        raise TypeError(f"Unknown type: {type}")

#--------------------------------------------------
# Constants
#--------------------------------------------------

# map literal types to IR types
TYPE_MAPPING = {
    str: types.String,
    bool: types.Bool,
    int: types.Int,
    float: types.Float,
    datetime.date: types.Date,
    datetime.datetime: types.DateTime,
}

# map potential promotions to their cost
LATTICE:dict[tuple[ir.Type|None, ir.Type|None], float] = {
    (types.Int, types.Decimal): 1.0,
    (types.Decimal, types.Float): 2.0,
    (types.Int, types.Float): 3.0,
}

PRIMITIVES = [types.Int, types.Float, types.Decimal, types.Bool, types.String, types.Date, types.DateTime]

ENTITY_LIKES = [types.RowId]

# these are relations that try to preserve their input types
# e.g. if you add a USD and a Decimal, the result is still a USD
TYPE_PRESERVERS = [
    *builtins.plus.overloads,
    *builtins.minus.overloads,
    *builtins.mul.overloads,
    *builtins.div.overloads,
    *builtins.sum.overloads,
    *builtins.max.overloads,
    *builtins.min.overloads,
]

#--------------------------------------------------
# Type helpers
#--------------------------------------------------

def is_abstract(type:ir.Type) -> bool:
    return type in [types.Any, types.Number]

def to_type(value: ir.Value|ir.Field) -> ir.Type:
    if isinstance(value, (ir.Var, ir.Field, ir.Literal)):
        return value.type

    value_type = type(value)
    if value_type in TYPE_MAPPING:
        return TYPE_MAPPING[value_type]

    if isinstance(value, tuple):
        return types.AnyList

    raise TypeError(f"Cannot determine IR type for value: {value} of type {type(value).__name__}")

def type_matches(actual:ir.Type, expected:ir.Type, allow_expected_parents=False) -> bool:
    is_scalar = isinstance(actual, ir.ScalarType)
    exp_is_scalar = isinstance(expected, ir.ScalarType)
    if actual == expected:
        return True
    if is_scalar and any([type_matches(parent, expected) for parent in actual.super_types]):
        return True
    if allow_expected_parents and exp_is_scalar and any([type_matches(actual, parent) for parent in expected.super_types]):
        return True
    if not is_primitive(expected) and exp_is_scalar and actual in ENTITY_LIKES:
        return True
    if isinstance(actual, ir.UnionType):
        for t in actual.types:
            if type_matches(t, expected):
                return True
    if actual == types.Any or expected == types.Any:
        return True
    if actual == types.Number and expected in (types.Int, types.Float, types.Decimal):
        return True
    if expected == types.Number and actual in (types.Int, types.Float, types.Decimal):
        return True
    return False

def promotion_cost(actual:ir.Type, expected:ir.Type, arg_types:set[ir.Type]) -> float:
    if type_matches(actual, expected):
        return 0
    # if we have a type variable and all the types in the expression
    # match, then there's no promotion needed and this is a valid overload
    elif expected is types.EntityTypeVar and not is_primitive(actual) and len(arg_types) == 1:
        return 0
    base_actual = to_base_primitive(actual)
    base_expected = to_base_primitive(expected)
    if (base_actual, base_expected) in LATTICE:
        return LATTICE[(base_actual, base_expected)]
    return float("inf")

def merge_types(type1: ir.Type, type2: ir.Type) -> ir.Type:
    combined = ordered_set()
    types_to_process = [type1, type2]

    # Iterative flattening of union types
    while types_to_process:
        t = types_to_process.pop()
        if isinstance(t, ir.UnionType):
            types_to_process.extend(t.types)
        else:
            combined.add(t)

    # If we have multiple types and Any is one of them, remove Any
    if len(combined) > 1 and types.Any in combined:
        combined.remove(types.Any)

    # Return single type or create a union
    return next(iter(combined)) if len(combined) == 1 else f.union_type(list(combined))

def to_base_primitive(type:ir.Type) -> Optional[ir.Type]:
    if isinstance(type, ir.ScalarType):
        if is_base_primitive(type):
            return type
        # walk the hierarchy to find a base primitive
        for parent in type.super_types:
            if found := to_base_primitive(parent):
                return found
    if isinstance(type, ir.UnionType):
        for t in type.types:
            if found := to_base_primitive(t):
                return found
    return None

def is_base_primitive(type:ir.Type) -> bool:
    return type in PRIMITIVES

def is_primitive(type:ir.Type) -> bool:
    return to_base_primitive(type) is not None

def invalid_type(type:ir.Type) -> bool:
    if isinstance(type, ir.UnionType):
        # if there are multiple primitives, or a primtive and a non-primitive
        # then we have an invalid type
        if len(type.types) > 1:
            return any([is_primitive(t) for t in type.types])
    return False

def try_preserve_type(types:set[ir.Type]) -> Optional[ir.Type]:
    # we keep the input type as the output type if either all inputs
    # are the exact same type or there's one nominal and its base primitive
    # type, e.g. USD + Decimal
    if len(types) == 1:
        return next(iter(types))
    if len(types) == 2:
        t1, t2 = types
        base_equivalent = type_matches(t1, t2, allow_expected_parents=True)
        if base_equivalent:
            # as long as one of the types is a base primitive, we can use the
            # other type as final preserved type
            if is_base_primitive(t1):
                return t2
            elif is_base_primitive(t2):
                return t1
    return None

#--------------------------------------------------
# Type Errors
#--------------------------------------------------

@dataclass
class TyperError():
    node: ir.Node

@dataclass
class TypeMismatch(TyperError):
    expected: ir.Type
    actual: ir.Type
    def __str__(self):
        return f"[red bold][Type Mismatch][/red bold] expected [yellow]{to_name(self.expected)}[/yellow], got [red]{to_name(self.actual)}[/red]: [white]{str(self.node).strip()}[/white]"

@dataclass
class InvalidType(TyperError):
    type: ir.Type
    def __str__(self):
        return f"[red bold][Invalid Type][/red bold] incompatible types [red]{to_name(self.type)}[/red]: [white]{str(self.node).strip()}[/white]"

@dataclass
class UnresolvedOverload(TyperError):
    arg_types: list[ir.Type]
    def __str__(self):
        assert isinstance(self.node, (ir.Lookup, ir.Update, ir.Aggregate))
        rel = to_relation(self.node)
        types = ', '.join([to_name(t) for t in self.arg_types])
        return f"[red bold][Unresolved Overload][/red bold] [yellow]{rel.name}({types})[/yellow]: [white]{str(self.node).strip()}[/white]"

@dataclass
class UnresolvedType(TyperError):
    def __str__(self):
        return f"[red bold][Unresolved Type][/red bold] [white]{str(self.node).strip()}[/white]"
    pass

#--------------------------------------------------
# Propagation Network
#--------------------------------------------------

# The core idea of the typer is to build a propagation network where nodes
# are vars, fields, or overloaded lookups/updates/aggregates. The intuition
# is that _all_ types in the IR ultimately flow from relation fields, so if
# we figure those out we just need to propagate their types to unknown vars, which
# may then flow into other fields and so on.

# This means the network only needs to contain nodes that either directly flow into
# an abstract node or are themselves abstract. We need to track overloads because
# their arguments effectively act like abstract vars until we've resolved the final types.

Node = Union[ir.Var, ir.Field, ir.Lookup, ir.Update, ir.Aggregate]

class PropagationNetwork():
    def __init__(self):

        # track the set of nodes that represent entry points into the network
        self.roots = ordered_set()
        # we separately want to track nodes that were loaded from a previous run
        # so that even if we have edges to them, we _still_ consider them roots
        # and properly propagate types from them at the beginning
        self.loaded_roots = set()
        self.edges:dict[Node, OrderedSet[Node]] = defaultdict(lambda: ordered_set())
        self.has_incoming = set()
        self.type_requirements:dict[Node, OrderedSet[ir.Field]] = defaultdict(lambda: ordered_set())

        self.errors:list[TyperError] = []
        self.resolved_types:dict[Node, ir.Type] = {}
        self.resolved_overloads:dict[Node, list[ir.Relation]] = {}

    #--------------------------------------------------
    # Mismatches
    #--------------------------------------------------

    def mismatch(self, node:Node, expected:ir.Type, actual:ir.Type):
        self.errors.append(TypeMismatch(node, expected, actual))

    def invalid_type(self, node:Node, type:ir.Type):
        self.errors.append(InvalidType(node, type))

    def unresolved_overload(self, node:ir.Lookup|ir.Update|ir.Aggregate):
        self.errors.append(UnresolvedOverload(node, [self.resolve(a) for a in node.args]))

    def unresolved_type(self, node:Node):
        self.errors.append(UnresolvedType(node))

    def has_errors(self, node:Node) -> bool:
        for mismatch in self.errors:
            if mismatch.node == node:
                return True
        return False

    #--------------------------------------------------
    # Types and Edges
    #--------------------------------------------------

    def add_edge(self, source:Node, target:Node):
        if target not in self.loaded_roots:
            self.roots.remove(target)
        if source not in self.has_incoming:
            self.roots.add(source)
        self.edges[source].add(target)
        self.has_incoming.add(target)

    def add_type(self, node:Node, type:ir.Type):
        if node in self.resolved_types:
            self.resolved_types[node] = merge_types(self.resolved_types[node], type)
        else:
            self.resolved_types[node] = type

    def add_type_requirement(self, source:Node, field:ir.Field):
        self.type_requirements[source].add(field)

    #--------------------------------------------------
    # Load previous types
    #--------------------------------------------------

    def load_types(self, type_dict:dict[Node, ir.Type]):
        for node, type in type_dict.items():
            if isinstance(node, (ir.Field)):
                self.add_type(node, type)
                self.loaded_roots.add(node)
                self.roots.add(node)

    #--------------------------------------------------
    # Resolve
    #--------------------------------------------------

    def resolve(self, value:Node|ir.Value) -> ir.Type:
        if isinstance(value, (ir.Var, ir.Field)):
            return self.resolved_types.get(value) or to_type(value)
        assert not isinstance(value, (ir.Lookup, ir.Update, ir.Aggregate)), "Should never try to resolve a task"
        return to_type(value)

    #--------------------------------------------------
    # Overloads
    #--------------------------------------------------

    def resolve_overload_deps(self, op:ir.Lookup|ir.Update|ir.Aggregate, keep_known=False):
        # We need to find which args flow into this task, which flow out, and
        # if some are inputs, then they are required before we can resolve
        incoming:list[Node] = []
        outgoing:list[Node] = []
        required:list[Node] = []
        relation = to_relation(op)
        field_args = zip(relation.fields, [self.resolve(a) for a in op.args], op.args)
        has_inputs = any([field.input for field in relation.fields])
        for field, arg_type, arg in field_args:
            if (is_abstract(arg_type) or keep_known) and isinstance(arg, ir.Var):
                if field.input:
                    incoming.append(arg)
                    required.append(arg)
                elif not has_inputs:
                    incoming.append(arg)
                    outgoing.append(arg)
                else:
                    outgoing.append(arg)

        # check if the overloads of this relation are fully resolved, if any aren't
        # then their fields are incoming edges to this node
        for overload in relation.overloads:
            for field in overload.fields:
                if is_abstract(self.resolve(field)):
                    incoming.append(field)
                    required.append(field)

        return incoming, outgoing, required

    def resolve_overload(self, op:ir.Lookup|ir.Update|ir.Aggregate) -> Optional[list[ir.Relation]]:
        # check if we have any unresolved args that are required, if all of our args
        # are unresolved (len(incoming) == len(relation.fields)) then we have no information
        # to try and resolve the overloads with
        incoming, outgoing, required = self.resolve_overload_deps(op)
        relation = to_relation(op)
        if required or len(incoming) == len(relation.fields):
            return

        # otherwise we compute the cost of each overload and return the set of relations
        # that have the lowest cost. This can be multiple in e.g. the Person.pets.name
        # case where Cat_name and Dog_name are equally valid
        inf = float("inf")
        min_cost = inf
        matches = []
        for overload in relation.overloads:
            arg_types = set(self.resolve(arg) for arg, field in zip(op.args, overload.fields) if field.input)
            total = 0
            for arg, field in zip(op.args, overload.fields):
                arg_type = self.resolve(arg)
                field_type = self.resolve(field)
                total += promotion_cost(arg_type, field_type, arg_types)
            if total != inf and total <= min_cost:
                if total < min_cost:
                    min_cost = total
                    matches.clear()
                matches.append(overload)

        return matches

    #--------------------------------------------------
    # Propagation
    #--------------------------------------------------

    def propagate(self):
        edges = self.edges
        work_list = []

        # go through all the roots and find any that are not abstract, they'll
        # be the first nodes to push types through the network
        unhandled = set()
        for node in self.roots:
            if not isinstance(node, (ir.Var, ir.Field)):
                continue
            node_type = self.resolve(node)
            if not is_abstract(node_type):
                work_list.append(node)
            else:
                unhandled.add(node)

        # push known type nodes through the edges
        for source in work_list:
            unhandled.discard(source)
            source_type = self.resolve(source)
            # check to see if the source has ended up with a set of types that
            # aren't valid, e.g. a union of primitives
            if invalid_type(source_type):
                self.invalid_type(source, source_type)

            # propagate our type to each outgoing edge
            for out in edges.get(source, []):
                # if this is an overload then we need to try and resolve it
                if isinstance(out, (ir.Update, ir.Lookup, ir.Aggregate)):
                    found = self.resolve_overload(out)
                    if found is not None:
                        start_arg_types = [self.resolve(arg) for arg in out.args]
                        self.propagate_overloads(out, found)
                        final_arg_types = [self.resolve(arg) for arg in out.args]
                        for arg, start, final in zip(out.args, start_arg_types, final_arg_types):
                            if start != final:
                                work_list.append(arg)
                # otherwiser we just add to the outgoing node's type and if it
                # changes we add it to the work list
                elif start := self.resolve(out):
                    self.add_type(out, source_type)
                    if start != self.resolve(out) or out in unhandled:
                        work_list.append(out)

        for source in unhandled:
            self.unresolved_type(source)

        # now that we've pushed all the types through the network, we need to validate
        # that all type requirements of those nodes are met
        for node, fields in self.type_requirements.items():
            node_type = self.resolve(node)
            for field in fields:
                field_type = self.resolve(field)
                if not type_matches(node_type, field_type):
                    self.mismatch(node, field_type, node_type)

        return self.resolved_types


    def propagate_overloads(self, node:ir.Lookup|ir.Update|ir.Aggregate, overloads:list[ir.Relation]):
        if not overloads:
            return self.unresolved_overload(node)

        # we've resolved the overloads, so store that
        self.resolved_overloads[node] = overloads

        # we need to determine the final types of our args but taking all the overloads
        # and adding the type of their fields back to the args.
        resolved_args = [self.resolve(a) for a in node.args]
        for overload in overloads:
            resolved_fields = [self.resolve(f) for f in overload.fields]
            # if our overload preserves types, we check to see if there's a preserved
            # output type given the inputs and if so, shadow the field's type with the
            # preserved type
            if overload in TYPE_PRESERVERS:
                input_types = set([arg_type for field, arg_type
                                    in zip(overload.fields, resolved_args)
                                    if field.input])
                if out_type := try_preserve_type(input_types):
                    resolved_fields = [field_type if field.input else out_type
                                        for field, field_type in zip(overload.fields, resolved_fields)]
            for field_type, arg_type, arg in zip(resolved_fields, resolved_args, node.args):
                if is_abstract(arg_type) and isinstance(arg, ir.Var):
                    self.add_type(arg, field_type)



    #--------------------------------------------------
    # Display
    #--------------------------------------------------

    # draw the network as a mermaid graph for the debugger
    def to_mermaid(self, max_edges=500) -> str:
        resolved = self.resolved_types
        nodes = ordered_set()
        link_strs = []
        for src, dsts in self.edges.items():
            nodes.add(src)
            for dst in dsts:
                if len(link_strs) > max_edges:
                    break
                nodes.add(dst)
                link_strs.append(f"n{src.id} --> n{dst.id}")
            if len(link_strs) > max_edges:
                break

        for src, dsts in self.type_requirements.items():
            nodes.add(src)
            for dst in dsts:
                if len(link_strs) > max_edges:
                    break
                nodes.add(dst)
                link_strs.append(f"n{src.id} --> n{dst.id}")
            if len(link_strs) > max_edges:
                break

        def type_span(t:ir.Type) -> str:
            type_str = t.name if isinstance(t, ir.ScalarType) else str(t)
            return f"<span style='color:cyan;'>{type_str.strip()}</span>"

        def overload_span(rel:ir.Relation, arg_types:list[ir.Type]) -> str:
            args = []
            for field, arg_type in zip(rel.fields, arg_types):
                field_type = self.resolve(field)
                if not type_matches(arg_type, field_type) and field_type != types.EntityTypeVar:
                    args.append(f"<span style='color:yellow;'>{str(arg_type).strip()} -> {str(field_type).strip()}</span>")
                elif isinstance(arg_type, ir.UnionType):
                    args.append(type_span(field_type))
                else:
                    args.append(type_span(arg_type))
            return f'{rel.name}({", ".join(args)})'

        node_strs = []
        for node in nodes:
            klass = ""
            if isinstance(node, ir.Var):
                ir_type = resolved.get(node) or self.resolve(node)
                type_str = type_span(ir_type)
                # if this is one of our vars that is a placeholder for a literal, just use the type
                if isinstance(ir_type, ir.ScalarType) and node.name == ir_type.name:
                    label = f'(["Literal {type_str}"])'
                else:
                    label = f'(["{node.name}: {type_str}"])'
            elif isinstance(node, ir.Field):
                ir_type = resolved.get(node) or self.resolve(node)
                type_str = type_span(ir_type)
                klass = ":::field"
                label = f"{{{{\"{node.name}: {type_str}\"}}}}"
            elif isinstance(node, (ir.Lookup, ir.Update, ir.Aggregate)):
                arg_types = [self.resolve(arg) for arg in node.args]
                if node in self.resolved_overloads:
                    overloads = self.resolved_overloads[node]
                    content = "<br/>".join([overload_span(o, arg_types) for o in overloads])
                else:
                    content = overload_span(to_relation(node), arg_types)
                label = f'[/"{content}"/]'
            else:
                raise NotImplementedError(f"Unknown node type: {type(node)}")
            if self.has_errors(node):
                klass = ":::error"
            node_strs.append(f"n{node.id}{label}{klass}")

        node_str = "\n                ".join(node_strs)
        link_str = "\n                ".join(link_strs)
        template = f"""
            %%{{init: {{'theme':'dark', 'flowchart':{{'useMaxWidth':false, 'htmlLabels': true}}}}}}%%
            flowchart TD
                linkStyle default stroke:#666
                classDef field fill:#245,stroke:#478
                classDef error fill:#624,stroke:#945,color:#f9a
                classDef default stroke:#444,stroke-width:2px, font-size:12px

                %% nodes
                {node_str}

                %% edges
                {link_str}
        """
        return template

    # simplified, less verbose (compared to mermaid) output for snapshot testing
    def to_fish(self) -> str:
        resolved = self.resolved_types
        nodes = ordered_set()
        for src, dsts in self.edges.items():
            nodes.add(src)
            nodes.update(dsts)
        for src, dsts in self.type_requirements.items():
            nodes.add(src)
            nodes.update(dsts)

        def type_to_str(t:ir.Type) -> str:
            type_str = t.name if isinstance(t, ir.ScalarType) else str(t)
            return type_str.strip()

        def overload_str(rel:ir.Relation, arg_types:list[ir.Type]) -> str:
            args = []
            for field, arg_type in zip(rel.fields, arg_types):
                field_type = self.resolve(field)
                if not type_matches(arg_type, field_type):
                    args.append(f"{str(arg_type).strip()} -?> {str(field_type).strip()}")
                elif isinstance(arg_type, ir.UnionType):
                    args.append(type_to_str(field_type))
                else:
                    args.append(type_to_str(arg_type))
            return f'{rel.name}({", ".join(args)})'

        def node_kind(node:ir.Node) -> str:
            if isinstance(node, (ir.Lookup, ir.Update, ir.Aggregate)):
                return "overload"
            return type(node).__name__.lower()

        def error_info(err:TyperError) -> str:
            if isinstance(err, TypeMismatch):
                return f'Type Mismatch|expected {to_name(err.expected)}, got {to_name(err.actual)}'
            return f'{type(err).__name__}'

        nl = "\n"
        node_strs = []
        for node in nodes:
            info = ""
            if isinstance(node, ir.Var):
                ir_type = resolved.get(node) or self.resolve(node)
                if not(isinstance(ir_type, ir.ScalarType) and node.name == ir_type.name):
                    info = f'{node.name}|{type_to_str(ir_type)}'
            elif isinstance(node, ir.Field):
                ir_type = resolved.get(node) or self.resolve(node)
                info = f'{node.name}|{type_to_str(ir_type)}'
            elif isinstance(node, (ir.Lookup, ir.Update, ir.Aggregate)):
                arg_types = [self.resolve(arg) for arg in node.args]
                if node in self.resolved_overloads:
                    overloads = self.resolved_overloads[node]
                    content = nl.join([overload_str(o, arg_types) for o in overloads])
                else:
                    content = overload_str(to_relation(node), arg_types)
                info = f'{content}'
            else:
                raise NotImplementedError(f"Unknown node type: {type(node)}")

            if info:
                error_suffix = " !" if self.has_errors(node) else ""
                node_strs.append(f'{node_kind(node)}|{info}{error_suffix}')

        node_strs.sort()
        if self.errors:
            node_strs.append("---ERRORS---")
            for err in self.errors:
                node_strs.append(f'{error_info(err)} ({str(err.node).strip()})')

        return f"""{nl.join(node_strs)}"""

#--------------------------------------------------
# Analyzer
#--------------------------------------------------

class Analyzer(visitor.Visitor):
    def __init__(self):
        super().__init__()
        self.net = PropagationNetwork()

        # this is a map of literal types to a var representing that literal in
        # the graph. This allows us to collapse all literals into a single node
        # and avoids exploding node count if lots of data gets added directly
        self.shared_literal_vars:dict[ir.Type, ir.Var] = {
            t: f.var(t.name, t)
            for t in types.builtin_types
            if isinstance(t, ir.ScalarType)
        }

    #--------------------------------------------------
    # Literal types
    #--------------------------------------------------

    # given a literal type, lookup or create a var for it
    def shared_literal_var(self, node:ir.Type) -> ir.Var:
        assert isinstance(node, ir.ScalarType)
        if node not in self.shared_literal_vars:
            self.shared_literal_vars[node] = f.var(node.name, node)
        return self.shared_literal_vars[node]

    #--------------------------------------------------
    # Lookups + Aggregates
    #--------------------------------------------------

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]):
        self.visit_rel_op(node, parent)

    def visit_aggregate(self, node: ir.Aggregate, parent: ir.Node | None):
        self.visit_rel_op(node, parent)

    def visit_rel_op(self, node: ir.Lookup|ir.Aggregate, parent: Optional[ir.Node]):
        rel = to_relation(node)
        if isinstance(node, ir.Lookup) and rel == builtins.eq:
            return self.visit_eq(node, parent)

        if rel.overloads:
            return self.visit_overloaded(node, parent)

        # if this is a population check, then it's fine to pass a subtype in to do the check
        # e.g. Employee(Person) is a valid way to check if a person is an employee
        allow_any_parents = builtins.concept_relation_annotation in rel.annotations
        for arg, field in zip(node.args, rel.fields):
            field_type = field.type
            arg_type = self.net.resolve(arg)
            if not type_matches(arg_type, field_type, allow_any_parents):
                self.net.mismatch(node, field_type, arg_type)
            # if we have an abstract var then this field will ultimately propagate to that
            # var's type
            elif isinstance(arg, ir.Var) and is_abstract(arg_type) and not field.input:
                self.net.add_edge(field, arg)

    #--------------------------------------------------
    # Overloads
    #--------------------------------------------------

    def visit_overloaded(self, node: ir.Lookup|ir.Update|ir.Aggregate, parent: Optional[ir.Node]):
        incoming, outgoing, required = self.net.resolve_overload_deps(node, keep_known=True)
        for arg in incoming:
            self.net.add_edge(arg, node)
        for arg in outgoing:
            self.net.add_edge(node, arg)

    #--------------------------------------------------
    # Eq
    #--------------------------------------------------

    def visit_eq(self, node: ir.Lookup, parent: Optional[ir.Node]):
        (left, right) = node.args
        left_type = self.net.resolve(left)
        right_type = self.net.resolve(right)
        if is_abstract(left_type) and is_abstract(right_type):
            assert isinstance(left, ir.Var) and isinstance(right, ir.Var)
            # if both sides are abstract, then whatever we find out about
            # either should propagate to the other
            self.net.add_edge(left, right)
            self.net.add_edge(right, left)
        elif is_abstract(left_type) and isinstance(right, ir.Var):
            assert isinstance(left, ir.Var)
            self.net.add_edge(right, left)
        elif is_abstract(right_type) and isinstance(left, ir.Var):
            assert isinstance(right, ir.Var)
            self.net.add_edge(left, right)
        elif not type_matches(left_type, right_type):
            self.net.mismatch(node, left_type, right_type)

    #--------------------------------------------------
    # Update
    #--------------------------------------------------

    def visit_update(self, node: ir.Update, parent: Optional[ir.Node]):
        rel = node.relation
        # if this is a population check, then it's fine to pass a subtype in to do the population
        # e.g. Employee(Person) should be a valid way to populate a person
        allow_any_parents = builtins.concept_relation_annotation in rel.annotations
        for arg, field in zip(node.args, rel.fields):
            field_type = field.type
            arg_type = self.net.resolve(arg)
            is_var = isinstance(arg, ir.Var)

            # if the arg is abstract, but the field isn't, then we need to make
            # sure that once the arg is resolved we check that it matches the field
            # type
            if is_var and is_abstract(arg_type) and not is_abstract(field_type):
                self.net.add_type_requirement(arg, field)

            # if the field is abstract, then eventually this var will help determine
            # the field's type
            if is_abstract(field_type):
                source = self.shared_literal_var(arg_type) if not is_var else arg
                assert source
                self.net.add_edge(source, field)

            elif not type_matches(arg_type, field_type, allow_expected_parents=allow_any_parents):
                self.net.mismatch(node, field_type, arg_type)

#--------------------------------------------------
# Replacer
#--------------------------------------------------

conversions:dict[tuple[ir.Type|None, ir.Type|None], ir.Relation] = {
    (types.Int, types.Float): builtins.int_to_float,
    (types.Int, types.Decimal): builtins.int_to_decimal,
    (types.Decimal, types.Float): builtins.decimal_to_float,
}

# Once we've pushed all the types through the network, we need to
# to replace the types of fields and vars that we may have discovered.
# We also need to replace overloaded lookups with the choosen overloads
# and do any conversions that are needed.
class Replacer(visitor.Rewriter):
    def __init__(self, net:PropagationNetwork):
        super().__init__()
        self.net = net

    def handle_field(self, node: ir.Field, parent: ir.Node):
        if node in self.net.resolved_types:
            return f.field(node.name, self.net.resolved_types[node], node.input)
        return node

    def handle_var(self, node: ir.Var, parent: ir.Node):
        if node in self.net.resolved_types:
            return f.var(node.name, self.net.resolved_types[node])
        return node

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node):
        if node not in self.net.resolved_overloads:
            return super().handle_lookup(node, parent)

        lookup_args = self.walk_list(node.args, node)

        # we can resolve to multiple overloads, that we'd need to union
        # so we'll need to hoist all the vars being used to make sure
        # everything joins correctly
        resolved = self.net.resolved_overloads[node]
        hoists = [arg for arg in lookup_args if isinstance(arg, ir.Var)]
        branches:list[ir.Logical] = []
        for overload in resolved:
            # we may do conversions, so we can end up with multiple tasks
            # in this branch and we need to track what the final args are
            tasks = []
            final_args = []
            for arg, field in zip(lookup_args, overload.fields):
                arg_type = to_type(arg)
                field_type = to_type(field)
                # the typer previously made sure that this should be valid so
                # a type mismatch means we need to convert as long as this isn't
                # a type variable
                if field_type != types.EntityTypeVar and not type_matches(to_type(arg), field_type):
                    name = f"v{to_name(field_type)}"
                    if isinstance(arg, ir.Var):
                        name = arg.name + "_" + to_name(field_type)
                    new_arg = f.var(name, field.type)
                    tasks.append(f.lookup(
                        conversions[(to_base_primitive(arg_type), to_base_primitive(field_type))],
                        [arg, new_arg]
                    ))
                    final_args.append(new_arg)
                else:
                    final_args.append(arg)
            tasks.append(f.lookup(overload, final_args))
            branches.append(f.logical(tasks, hoists))

        # unwrap if we don't actually need a union
        if len(branches) == 1:
            return branches[0].body
        else:
            return f.union(branches, hoists)

#--------------------------------------------------
# Typer pass
#--------------------------------------------------

class InferTypes(compiler.Pass):
    def __init__(self):
        super().__init__()
        self.historical = {}

    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        w = Analyzer()
        w.net.load_types(self.historical)

        # build the propagation network
        with debugging.span("type.analyze"):
            model.root.accept(w, model)

        # propagate the types through the network
        with debugging.span("type.propagate") as end_span:
            field_types = w.net.propagate()
            self.historical.update(field_types)
            end_span["type_graph"] = w.net.to_mermaid()
            end_span["type_report"] = w.net.to_fish()

        # replace the fields in the model with the new types
        with debugging.span("type.replace"):
            final = Replacer(w.net).walk(model)

        for err in w.net.errors:
            rich.print(str(err))

        return final



from relationalai.early_access.lqp.validators import assert_valid_input
from relationalai.early_access.metamodel import ir, builtins as rel_builtins, helpers, types
from relationalai.early_access.metamodel.visitor import collect_by_type
from relationalai.early_access.lqp import ir as lqp
from relationalai.early_access.lqp.hash_utils import lqp_hash
from relationalai.early_access.lqp.primitives import relname_to_lqp_name, lqp_operator, lqp_avg_op
from relationalai.early_access.lqp.types import meta_type_to_lqp, type_from_constant
from relationalai.early_access.lqp.constructors import mk_and, mk_exists, mk_or, mk_abstraction
from relationalai.early_access.lqp.utils import TranslationCtx, gen_unique_var

from typing import Tuple, cast, Union

""" Main access point. Converts the model IR to an LQP program. """
def to_lqp(model: ir.Model) -> lqp.LqpProgram:
    assert_valid_input(model)
    ctx = TranslationCtx(model)
    program = _translate_to_program(ctx, model)
    return program

def _translate_to_program(ctx: TranslationCtx, model: ir.Model) -> lqp.LqpProgram:
    decls: list[lqp.Declaration] = []
    outputs: list[Tuple[str, lqp.RelationId]] = []

    # LQP only accepts logical tasks
    # These are asserted at init time
    root = cast(ir.Logical, model.root)

    seen_rids = set()
    for subtask in root.body:
        assert isinstance(subtask, ir.Logical)
        new_decls = _translate_to_decls(ctx, subtask)
        for decl in new_decls:
            assert isinstance(decl, lqp.Def), f"expected Def, got {type(decl)}: {decl}"
            rid = decl.name
            assert rid not in seen_rids, f"duplicate relation id: {rid}"
            seen_rids.add(rid)

            decls.append(decl)

    for output_id in ctx.output_ids:
        assert isinstance(output_id, lqp.RelationId)
        outputs.append(("output", output_id))

    debug_info = lqp.DebugInfo(id_to_orig_name=ctx.id_to_orig_name)

    return lqp.LqpProgram(defs=decls, outputs=outputs, debug_info=debug_info)

def _effect_bindings(effect: Union[ir.Output, ir.Update]) -> list[ir.Value]:
    if isinstance(effect, ir.Output):
        # Unions may not return anything. The generated IR contains a None value when this
        # happens. We ignore it here.
        # TODO: Improve handling of empty union outputs
        # TODO: we dont yet handle aliases, so we ignore v[0]
        return [v[1] for v in effect.aliases if v[1] is not None]
    else:
        return list(effect.args)

def _translate_to_decls(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp.Declaration]:
    effects = collect_by_type((ir.Output, ir.Update), rule)
    aggregates = collect_by_type(ir.Aggregate, rule)
    ranks = collect_by_type(ir.Rank, rule)

    # TODO: should this ever actually come in as input?
    if len(effects) == 0:
        return []

    assert len(ranks) == 0 or len(aggregates) == 0, "rules cannot have both aggregates and ranks"

    conjuncts = []
    for task in rule.body:
        if isinstance(task, (ir.Output, ir.Update)):
            continue
        conjuncts.append(_translate_to_formula(ctx, task))

    # Aggregates reduce over the body
    if len(aggregates) > 0:
        aggr_body = mk_and(conjuncts)
        conjuncts = []
        for aggr in aggregates:
            conjuncts.append(_translate_aggregate(ctx, aggr, aggr_body))

    if len(ranks) > 0:
        rank_body = mk_and(conjuncts)
        conjuncts = []
        for rank in ranks:
            conjuncts.append(_translate_rank(ctx, rank, rank_body))

    return [_translate_effect(ctx, effect, mk_and(conjuncts)) for effect in effects]

def _translate_effect(ctx: TranslationCtx, effect: Union[ir.Output, ir.Update], body: lqp.Formula) -> lqp.Declaration:
    # Handle the bindings
    bindings = _effect_bindings(effect)
    projection, eqs = translate_bindings(ctx, bindings)
    eqs.append(body)
    new_body = mk_and(eqs)

    is_output = isinstance(effect, ir.Output)
    def_name = "output" if is_output else effect.relation.name
    meta_id = effect.id if is_output else effect.relation.id
    rel_id = get_relation_id(ctx, def_name, meta_id)

    # Context bookkeeping
    if is_output:
        ctx.output_ids.append(rel_id)

    # TODO: is this correct? might need attrs tooo?
    return lqp.Def(
        name = rel_id,
        body = mk_abstraction(projection, new_body),
        attrs = [],
        meta = None,
    )

def _translate_rank(ctx: TranslationCtx, rank: ir.Rank, body: lqp.Formula) -> lqp.Formula:
    # Get direction of ranking, then replicate std::common::sort or
    # std::common::reverse_sort

    # TODO: handle descending rank
    # if  all(o for o in rank.arg_is_ascending):
    #     raise NotImplementedError("Descending rank direction is not yet supported")
    # TODO: handle multiple rank args
    # TODO: Handle limits
    # TODO: Do it correctly

    input_args, input_eqs = translate_bindings(ctx, list(rank.args[:-1]))
    output_var, _ = _translate_term(ctx, rank.args[-1])
    projected_args, projected_eqs = translate_bindings(ctx, list(rank.projection))

    body_conjs = [body]
    body_conjs.extend(input_eqs)
    body_conjs.extend(projected_eqs)
    body = mk_and(body_conjs)

    abstr_args: list[Tuple[lqp.Var, lqp.RelType]] = projected_args + input_args
    result_var, _ = _translate_term(ctx, rank.result)
    terms = [result_var] + [t for t,_ in input_args] + [output_var]

    return lqp.FFI(
        meta=None,
        name="rel_primitive_sort",
        args=[mk_abstraction(abstr_args, body)],
        terms=terms,
    )


def _translate_aggregate(ctx: TranslationCtx, aggr: ir.Aggregate, body: lqp.Formula) -> Union[lqp.Reduce, lqp.Formula]:
    # TODO: handle this properly
    aggr_name = aggr.aggregation.name
    supported_aggrs = ("sum", "count", "avg", "min", "max", "rel_primitive_solverlib_ho_appl")
    assert aggr_name in supported_aggrs, f"only support {supported_aggrs} for now, not {aggr.aggregation.name}"

    meta_output_terms = []
    meta_input_terms = []

    for (field, arg) in zip(aggr.aggregation.fields, aggr.args):
        if field.input:
            meta_input_terms.append(arg)
        else:
            meta_output_terms.append(arg)

    output_vars = [_translate_term(ctx, term)[0] for term in meta_output_terms]

    body_conjs = [body]
    input_args, input_eqs = translate_bindings(ctx, meta_input_terms)
    projected_args, projected_eqs = translate_bindings(ctx, list(aggr.projection))
    body_conjs.extend(input_eqs)
    body_conjs.extend(projected_eqs)
    abstr_args: list[Tuple[lqp.Var, lqp.RelType]] = projected_args + input_args

    if aggr_name == "count" or aggr_name == "avg":
        # Count sums up "1"
        one_var, typ, eq = binding_to_lqp_var(ctx, 1)
        assert eq is not None
        body_conjs.append(eq)
        abstr_args.append((one_var, typ))

    body = mk_and(body_conjs)

    # Average needs to wrap the reduce in Exists(Conjunction(Reduce, div))
    if aggr_name == "avg":
        assert len(output_vars) == 1, "avg should only have one output variable"
        output_var = output_vars[0]

        # The average will produce two output variables: sum and count.
        sum_result = gen_unique_var(ctx, "sum")
        count_result = gen_unique_var(ctx, "count")

        # Second to last is the variable we're summing over.
        (sum_var, sum_type) = abstr_args[-2]

        result = lqp.Reduce(
            op=lqp_avg_op(ctx.unique_names, aggr.aggregation, sum_var.name, sum_type),
            body=mk_abstraction(abstr_args, body),
            terms=[sum_result, count_result],
            meta=None,
        )

        div = lqp.Primitive(name="rel_primitive_divide", terms=[sum_result, count_result, output_var], meta=None)
        conjunction = mk_and([result, div])

        # Finally, we need to wrap everything in an `exists` to project away the sum and
        # count variables and only keep the result of the division.
        result = mk_exists([(sum_result, sum_type), (count_result, lqp.PrimitiveType.INT)], conjunction)

        return result

    # `input_args`` hold the types of the input arguments, but they may have been modified
    # if we're dealing with a count, so we use `abstr_args` to find the type.
    (aggr_arg, aggr_arg_type) = abstr_args[-1]
    # Group-bys do not need to be handled at all, since they are introduced outside already
    reduce = lqp.Reduce(
        op=lqp_operator(ctx.unique_names, aggr.aggregation, aggr_arg.name, aggr_arg_type),
        body=mk_abstraction(abstr_args, body),
        terms=output_vars,
        meta=None
    )
    return reduce

def _translate_to_formula(ctx: TranslationCtx, task: ir.Task) -> lqp.Formula:
    if isinstance(task, ir.Logical):
        conjuncts = [_translate_to_formula(ctx, child) for child in task.body]
        return mk_and(conjuncts)
    elif isinstance(task, ir.Lookup):
        return _translate_to_atom(ctx, task)
    elif isinstance(task, ir.Not):
        return lqp.Not(arg=_translate_to_formula(ctx, task.task), meta=None)
    elif isinstance(task, ir.Exists):
        lqp_vars, conjuncts = translate_bindings(ctx, list(task.vars))
        conjuncts.append(_translate_to_formula(ctx, task.task))
        return mk_exists(lqp_vars, mk_and(conjuncts))
    elif isinstance(task, ir.Construct):
        assert len(task.values) >= 1, "construct should have at least one value"
        # TODO: what does the first value do
        terms = [_translate_term(ctx, arg) for arg in task.values[1:]]
        terms.append(_translate_term(ctx, task.id_var))
        return lqp.Primitive(
            name="rel_primitive_hash_tuple_uint128",
            terms=[v for v, _ in terms],
            meta=None
        )
    elif isinstance(task, ir.Union):
        # TODO: handle hoisted vars if needed
        disjs = [_translate_to_formula(ctx, child) for child in task.tasks]
        return mk_or(disjs)
    elif isinstance(task, (ir.Aggregate, ir.Output, ir.Update)):
        # Nothing to do here, handled in _translate_to_decls
        return mk_and([])
    elif isinstance(task, ir.Rank):
        # Nothing to do here, handled in _translate_to_decls
        return mk_and([])
    else:
        raise NotImplementedError(f"Unknown task type (formula): {type(task)}")

# Only used for translating terms on atoms, which can be specialized values.
def _translate_relterm(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.RelTerm, lqp.RelType]:
    if isinstance(term, ir.Literal) and term.type == types.Symbol:
        assert isinstance(term.value, str)
        return lqp.Specialized(value=term.value, meta=None), type_from_constant(term.value)

    return _translate_term(ctx, term)

def _translate_term(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.Term, lqp.RelType]:
    if isinstance(term, ir.Var):
        name = ctx.unique_names.get_name_by_id(term.id, term.name)
        t = meta_type_to_lqp(term.type)
        return lqp.var(name), t
    elif isinstance(term, ir.Literal):
        assert isinstance(term.value, lqp.PrimitiveValue)
        return term.value, type_from_constant(term.value)
    else:
        assert isinstance(term, lqp.PrimitiveValue), \
            f"Cannot translate value {term!r} of type {type(term)} to LQP Term; not a PrimitiveValue."
        return term, type_from_constant(term)

# In the metamodel, type conversions are represented as special relations, whereas in LQP we
# have a dedicated `Cast` node. Eventually we might want to unify these, but for now we use
# this mapping here.
rel_to_cast = {
    "int_to_float": lqp.PrimitiveType.FLOAT,
    "int_to_decimal": lqp.RelValueType.DECIMAL,
    "decimal_to_float": lqp.PrimitiveType.FLOAT,
}

def _translate_to_atom(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
    # TODO: want signature not name
    rel_name = task.relation.name
    terms = []
    for arg in task.args:
        # Handle varargs, which come wrapped in a tuple.
        if isinstance(arg, tuple):
            for vararg in arg:
                term, _t = _translate_relterm(ctx, vararg)
                terms.append(term)
        else:
            term, _t = _translate_relterm(ctx, arg)
            terms.append(term)

    if rel_builtins.is_builtin(task.relation):
        if task.relation.name in rel_to_cast:
            assert len(terms) == 2, f"expected two terms for cast {task.relation.name}, got {terms}"
            return lqp.Cast(type=rel_to_cast[task.relation.name], input=terms[0], result=terms[1], meta=None)
        else:
            lqp_name = relname_to_lqp_name(task.relation.name)
            return lqp.Primitive(name=lqp_name, terms=terms, meta=None)

    if helpers.is_external(task.relation):
        return lqp.RelAtom(name=task.relation.name, terms=terms, meta=None)

    rid = get_relation_id(ctx, rel_name, task.relation.id)
    return lqp.Atom(name=rid, terms=terms, meta=None)

def get_relation_id(ctx: TranslationCtx, orig_name: str, metamodel_id: int) -> lqp.RelationId:
    mid_str = str(metamodel_id)
    relation_id = lqp.RelationId(id=lqp_hash(mid_str), meta=None)
    ctx.id_to_orig_name[relation_id] = orig_name
    return relation_id

def translate_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.RelType]], list[lqp.Formula]]:
    lqp_vars = []
    conjuncts = []
    for binding in bindings:
        lqp_var, typ, eq = binding_to_lqp_var(ctx, binding)
        lqp_vars.append((lqp_var, typ))
        if eq is not None:
            conjuncts.append(eq)

    return lqp_vars, conjuncts

def binding_to_lqp_var(ctx: TranslationCtx, binding: ir.Value) -> Tuple[lqp.Var, lqp.RelType, Union[None, lqp.Formula]]:
    if isinstance(binding, ir.Var):
        var, typ = _translate_term(ctx, binding)
        assert isinstance(var, lqp.Var)
        return var, typ, None
    else:
        # Constant in this case
        assert isinstance(binding, lqp.PrimitiveValue), f"expected primitive value, got {type(binding)}: {binding}"
        value, typ = _translate_term(ctx, binding)

        var = gen_unique_var(ctx, "cvar")
        eq = lqp.Primitive(name="rel_primitive_eq", terms=[var, value], meta=None)
        return var, typ, eq



from relationalai.early_access.metamodel.compiler import Pass
from relationalai.early_access.metamodel import ir, builtins as rel_builtins, factory as f, visitor
from relationalai.early_access.metamodel.typer.checker import Checker
from relationalai.early_access.metamodel.typer.typer2 import InferTypes
from typing import cast, Union
from relationalai.early_access.metamodel import types

from relationalai.early_access.metamodel.rewrite import Splinter, GarbageCollectNodes

# TODO: Move this into metamodel.rewrite
from relationalai.early_access.rel.rewrite import Flatten, QuantifyVars, ExtractCommon, CDC

import datetime

def lqp_passes() -> list[Pass]:
    return [
        Checker(),
        InferTypes(),
        GarbageCollectNodes(),
        CDC(),
        ExtractCommon(),
        Flatten(),
        QuantifyVars(),
        Splinter(), # Adds missing existentials + splits multi-headed rules into single rules
        UnifyDefinitions(),
        EliminateData(),  # Turns Data nodes into ordinary relations.
        EliminateValueTypeConstants(),
    ]

class UnifyDefinitions(Pass):
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        multidef_relations = self.get_multidef_relations(model)
        for relation in multidef_relations:
            model = self.rename_multidef(model, relation)
        return model

    def get_multidef_relations(self, model: ir.Model) -> set[ir.Relation]:
        seen = set()
        result = set()
        root = cast(ir.Logical, model.root)
        for task in root.body:
            task = cast(ir.Logical, task)
            for subtask in task.body:
                if isinstance(subtask, ir.Update):
                    assert subtask.effect == ir.Effect.derive, "only derive updates supported yet"
                    name = subtask.relation
                    if name.id in seen:
                        result.add(name)
                    seen.add(name.id)
        return result

    def rename_multidef(self, model: ir.Model, relation: ir.Relation) -> ir.Model:
        defname = relation.name
        root = cast(ir.Logical, model.root)
        new_subtasks = []
        new_relations = []
        total_ct = 0
        for subtask in root.body:
            subtask = cast(ir.Logical, subtask)
            new_subsubtasks: list[ir.Task] = []
            for subsubtask in subtask.body:
                if isinstance(subsubtask, ir.Update):
                    assert subsubtask.effect == ir.Effect.derive, "only derive updates supported yet"
                    name = subsubtask.relation
                    if name.id == relation.id:
                        total_ct += 1
                        # TODO: this needs to be unique btw (gensym)
                        new_name = f"{defname}_{total_ct}"
                        new_subsubtask = self.rename_relation(subsubtask, new_name)
                        new_subsubtasks.append(new_subsubtask)
                        new_relation = ir.Relation(
                            new_name,
                            relation.fields,
                            relation.requires,
                        )
                        new_relations.append(new_relation)
                    else:
                        new_subsubtasks.append(subsubtask)
                else:
                    new_subsubtasks.append(subsubtask)

            new_subtask = ir.Logical(
                subtask.engine,
                subtask.hoisted,
                tuple(new_subsubtasks),
            )
            new_subtasks.append(new_subtask)
        assert total_ct > 0, f"should have found at least one definition for {defname}"

        args = []
        for field in relation.fields:
            args.append(ir.Var(field.type, field.name))

        # Also add the new definition
        final_relation = ir.Relation(
            defname,
            relation.fields,
            relation.requires,
        )
        new_update = ir.Update(
            root.engine,
            final_relation,
            tuple(args),
            ir.Effect.derive,
        )

        logical_tasks = []
        lookups = []
        for new_relation in new_relations:
            new_lookup = ir.Lookup(
                root.engine,
                new_relation,
                tuple(args),
            )
            lookups.append(new_lookup)

        disj = ir.Union(
            root.engine,
            tuple(),
            tuple(lookups),
        )
        logical_tasks.append(disj)
        logical_tasks.append(new_update)
        new_logical = ir.Logical(
            root.engine,
            root.hoisted,
            tuple(logical_tasks),
        )
        new_subtasks.append(new_logical)

        new_root = ir.Logical(root.engine, root.hoisted, tuple(new_subtasks))
        model = ir.Model(
            model.engines,
            model.relations | new_relations,
            model.types,
            new_root,
        )
        return model

    def rename_relation(self, task: ir.Update, new_name: str) -> ir.Update:
        assert isinstance(task, ir.Update)
        assert task.effect == ir.Effect.derive, "only derive updates supported yet"
        old_relation = task.relation
        new_relation = ir.Relation(
            new_name,
            old_relation.fields,
            old_relation.requires,
        )
        new_task = ir.Update(
            task.engine,
            new_relation,
            task.args,
            task.effect,
        )
        return new_task

# We don't have a proto representation for these, so we rewrite them to be constructed directly
class EliminateValueTypeConstants(Pass):
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        p = RewriteBadLiterals()
        return p.walk(model)

def rewrite_bad_literals(self, args):
    vars_to_existify = []
    new_conjs = []
    new_args = []

    for arg in args:
        if isinstance(arg, datetime.datetime):
            new_var = f.var("dt_var", types.DateTime)
            new_args.append(new_var)
            vars_to_existify.append(new_var)

            year = arg.year
            month = arg.month
            day = arg.day
            hour = arg.hour
            minute = arg.minute
            second = arg.second

            lookup = f.lookup(
                    rel_builtins.construct_datetime,
                    tuple([
                        f.literal(year),
                        f.literal(month),
                        f.literal(day),
                        f.literal(hour),
                        f.literal(minute),
                        f.literal(second),
                        new_var,
                    ])
                )
            new_conjs.append(lookup)
        elif isinstance(arg, datetime.date):
            new_var = f.var("dt_var", types.Date)
            new_args.append(new_var)
            vars_to_existify.append(new_var)

            year = arg.year
            month = arg.month
            day = arg.day

            lookup = f.lookup(
                    rel_builtins.construct_date,
                    tuple([
                        f.literal(year),
                        f.literal(month),
                        f.literal(day),
                        new_var,
                    ])
                )
            new_conjs.append(lookup)
        else:
            new_args.append(arg)

    return new_args, vars_to_existify, new_conjs

class RewriteBadLiterals(visitor.Rewriter):
    def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> Union[ir.Lookup, ir.Exists]:
        args = node.args
        new_args, vars_to_existify, new_conjs = rewrite_bad_literals(self, args)

        if len(vars_to_existify) == 0:
            return node

        new_lookup = f.lookup(
            node.relation,
            tuple(new_args),
        )
        new_conjs.append(new_lookup)

        result = f.exists(
            vars_to_existify,
            f.logical(tuple(new_conjs)),
        )

        return result

    def handle_output(self, node: ir.Output, parent: ir.Node) -> Union[ir.Output, ir.Logical]:
        args = [v[1] for v in node.aliases]
        new_args, vars_to_existify, new_conjs = rewrite_bad_literals(self, args)

        if len(vars_to_existify) == 0:
            return node

        new_aliases = []
        for i, alias in enumerate(node.aliases):
            new_aliases.append((alias[0], new_args[i]))

        new_output = node.reconstruct(aliases=tuple(new_aliases))
        new_conjs.append(new_output)

        return f.logical(new_conjs)

    def handle_update(self, node: ir.Update, parent: ir.Node) -> Union[ir.Update, ir.Logical]:
        args = node.args
        new_args, vars_to_existify, new_conjs = rewrite_bad_literals(self, args)

        if len(vars_to_existify) == 0:
            return node

        new_update = node.reconstruct(args=tuple(new_args))
        new_conjs.append(new_update)
        return f.logical(new_conjs)

# Creates intermediary relations for all Data nodes and replaces said Data nodes
# with a Lookup into these created relations.
class EliminateData(Pass):
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        r = self.DataRewriter()
        return r.walk(model)

    # Does the actual work.
    class DataRewriter(visitor.Rewriter):
        new_relations: list[ir.Relation]
        new_updates: list[ir.Logical]
        # Counter for naming new relations.
        # It must be that new_count == len new_updates == len new_relations.
        new_count: int

        def __init__(self):
            self.new_relations = []
            self.new_updates = []
            self.new_count = 0
            super().__init__()

        # Create a new intermediary relation representing the Data (and pop it in
        # new_updates/new_relations) and replace this Data with a Lookup of said
        # intermediary.
        def handle_data(self, node: ir.Data, parent: ir.Node) -> ir.Lookup:
            self.new_count += 1
            intermediary_name = f"formerly_Data_{self.new_count}"

            intermediary_relation = f.relation(
                intermediary_name,
                [f.field(v.name, v.type) for v in node.vars]
            )
            self.new_relations.append(intermediary_relation)

            intermediary_update = f.logical([
                # For each row (union), equate values and their variable (logical).
                f.union(
                    [
                        f.logical(
                            [
                                f.lookup(rel_builtins.eq, [val, var])
                                for (val, var) in zip(row, node.vars)
                            ],
                            hoisted = node.vars,
                        )
                        for row in node
                    ],
                    hoisted = node.vars,
                ),
                # And pop it back into the relation.
                f.update(intermediary_relation, node.vars, ir.Effect.derive),
            ])
            self.new_updates.append(intermediary_update)

            replacement_lookup = f.lookup(intermediary_relation, node.vars)

            return replacement_lookup

        # Walks the model for the handle_data work then updates the model with
        # the new state.
        def handle_model(self, model: ir.Model, parent: None):
            walked_model = super().handle_model(model, parent)
            assert len(self.new_relations) == len(self.new_updates) and self.new_count == len(self.new_relations)

            # This is okay because its LQP.
            assert isinstance(walked_model.root, ir.Logical)
            root_logical = cast(ir.Logical, walked_model.root)

            # We may need to add the new intermediaries from handle_data to the model.
            if self.new_count  == 0:
                return model
            else:
                return ir.Model(
                    walked_model.engines,
                    walked_model.relations | self.new_relations,
                    walked_model.types,
                    ir.Logical(
                        root_logical.engine,
                        root_logical.hoisted,
                        root_logical.body + tuple(self.new_updates),
                        root_logical.annotations,
                    ),
                    walked_model.annotations,
                )



from __future__ import annotations
from dataclasses import dataclass
from typing import Union, Tuple

__all__ = [
    "SourceInfo",
    "LqpNode",
    "Declaration",
    "Def",
    "Loop",
    "Abstraction",
    "Formula",
    "Exists",
    "Reduce",
    "Conjunction",
    "Disjunction",
    "Not",
    "FFI",
    "Atom",
    "Pragma",
    "Primitive",
    "RelAtom",
    "Cast",
    "Var",
    "UInt128",
    "PrimitiveValue",
    "Constant",
    "Term",
    "Specialized",
    "RelTerm",
    "Attribute",
    "RelationId",
    "RelValueType",
    "PrimitiveType",
    "RelType",
    "FragmentId",
    "Fragment",
    "Define",
    "Write",
    "Output",
    "Read",
    "Epoch",
    "Transaction",
    "DebugInfo",
    "LqpProgram",
    "lqp_node_to_proto",
    "ir_to_proto",
    "convert_transaction",
    "var"
]

from lqp.ir import (
    SourceInfo,
    LqpNode,
    Declaration,
    Def,
    Loop,
    Abstraction,
    Formula,
    Exists,
    Reduce,
    Conjunction,
    Disjunction,
    Not,
    FFI,
    Atom,
    Pragma,
    Primitive,
    RelAtom,
    Cast,
    Var,
    UInt128,
    PrimitiveValue,
    Constant,
    Term,
    Specialized,
    RelTerm,
    Attribute,
    RelationId,
    RelValueType,
    PrimitiveType,
    RelType,
    FragmentId,
    Fragment,
    Define,
    Write,
    Output,
    Read,
    Epoch,
    Transaction
)

from lqp.emit import (
    ir_to_proto,
    convert_transaction
)

def var(name: str, meta: Union[SourceInfo, None] = None) -> Var:
    return Var(name=name, meta=meta)

@dataclass(frozen=True)
class DebugInfo:
    id_to_orig_name: dict[RelationId, str]

@dataclass(frozen=True)
class LqpProgram:
    defs: list[Declaration]
    outputs: list[Tuple[str, RelationId]]
    debug_info: Union[DebugInfo, None] = None

def lqp_program_to_transaction(program: LqpProgram, fragment_name: bytes = bytes(404)) -> Transaction:
    lqp_ir_fragment_id = FragmentId(id=fragment_name, meta=None)
    lqp_ir_fragment = Fragment(id=lqp_ir_fragment_id, declarations=program.defs, meta=None)
    lqp_ir_define_op = Define(fragment=lqp_ir_fragment, meta=None)
    lqp_ir_write = Write(write_type=lqp_ir_define_op, meta=None)
    lqp_ir_reads = []
    for name, rid in program.outputs:
        lqp_ir_output = Output(name=name, relation_id=rid, meta=None)
        lqp_ir_reads.append(Read(read_type=lqp_ir_output, meta=None))
    lqp_ir_epoch = Epoch(local_writes=[lqp_ir_write], reads=lqp_ir_reads, persistent_writes=[], meta=None)
    return Transaction(epochs=[lqp_ir_epoch], meta=None)

lqp_node_to_proto = ir_to_proto



from dataclasses import dataclass, field
from typing import Dict, Optional, Iterable, Union as PyUnion, Tuple, cast
from datetime import date, datetime
import networkx as nx

from relationalai.early_access.metamodel import ir, types, visitor, helpers
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set

# TODO: Simplify TypeVar so it does not contain constraint information
@dataclass
class TypeVar:
    """A type variable."""
    # The node that this type variable is bound to.
    # Here this is always a Var or a Field.
    # Together with the context and index, this uniquely identifies the type variable.
    node: PyUnion[ir.Field, ir.Var] = field(init=True)

    # When a type variable represents a tuple type, this is the index of the tuple element, or -1.
    index: int = field(init=True)

    # This is the enclosing lookup or aggregate node being applied.
    # The context is used to distinguish type variables for a Field at each use of the Field.
    context: PyUnion[ir.Lookup, ir.Aggregate, ir.Rank, None] = field(init=True)

    # The upper bound of the type variable.
    upper_bound: OrderedSet[ir.Type] = field(default_factory=OrderedSet, init=False, compare=False, hash=False, repr=False)
    # The lower bound of the type variable.
    lower_bound: OrderedSet[ir.Type] = field(default_factory=OrderedSet, init=False, compare=False, hash=False, repr=False)
    # Set of TypeVars that are supertypes of this type variable.
    node_super_types: OrderedSet["TypeVar"] = field(default_factory=OrderedSet, init=False, compare=False, hash=False, repr=False)
    # Inverse of node_super_types.
    node_sub_types: OrderedSet["TypeVar"] = field(default_factory=OrderedSet, init=False, compare=False, hash=False, repr=False)

    # Union-find data structure.
    # The rank of the type variable.
    rank: int = field(default=1, init=False, compare=False, hash=False, repr=False)
    # The next type variable in the union-find data structure. If this is None, the type variable is the root of its union-find tree.
    next: Optional["TypeVar"] = field(default=None, init=False, compare=False, hash=False, repr=False)

    def __hash__(self) -> int:
        return hash((self.node.id, self.index, self.context.id if self.context else None))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TypeVar):
            return False
        if self.node.id == other.node.id and self.index == other.index:
            if self.context and other.context:
                return self.context.id == other.context.id
            else:
                return self.context is None and other.context is None
        return False

    def __str__(self) -> str:
        base = TypeVar.pretty_name(self.node, self.index, self.context, lowercase=True)
        return f"{base} (id={self.node.id})"

    def find(self) -> "TypeVar":
        """Find the root of this type variable and perform path compression."""
        if self.next is None:
            return self
        self.next = self.next.find()
        return self.next

    def canonical_repr(self) -> "TypeVar":
        return self.find()

    def is_canonical(self) -> bool:
        return self.find() == self

    def union(self, other: "TypeVar") -> "TypeVar":
        """Union this type variable with another, returning the root of the union."""
        top = self.find()
        bot = other.find()
        assert top.next is None
        assert bot.next is None
        if top == bot:
            return top

        # But the lower rank tree is under the higher rank tree.
        if top.rank < bot.rank:
            top, bot = bot, top

        # For more consistent debugging output,
        # prefer that the root be a global field rather than a var,
        # and prefer global fields to instantiated fields.
        if isinstance(bot.node, ir.Field) and bot.context:
            top, bot = bot, top
        if isinstance(top.node, ir.Field) and not top.context:
            top, bot = bot, top

        # But bot under top.
        bot.next = top
        top.rank += bot.rank

        # Union bot's data into top.
        top.upper_bound.update(bot.upper_bound)
        top.lower_bound.update(bot.lower_bound)
        top.node_super_types.update(bot.node_super_types)
        top.node_sub_types.update(bot.node_sub_types)

        # Clear bot.
        bot.upper_bound.clear()
        bot.lower_bound.clear()
        bot.node_super_types.clear()
        bot.node_sub_types.clear()


        assert self.find() is top
        assert other.find() is top
        return top

    def compute_type(self) -> ir.Type:
        root = self.find()

        upper = types.compute_lowest_bound(root.upper_bound)
        lower = types.compute_highest_bound(root.lower_bound)

        if lower and upper:
            lower = types.intersect(lower, upper)

        if lower:
            return cast(ir.Type, lower)

        if upper:
            return cast(ir.Type, upper)

        return cast(ir.Type, types.Any)

    @staticmethod
    def pretty_name(node: ir.Node, index: int, context: Optional[ir.Node], lowercase: bool=False, short: bool=False) -> str:
        if short:
            if isinstance(node, ir.Var):
                return f"`{node.name}`"
            elif isinstance(node, ir.Field):
                if index < 0:
                    return f"`{node.name}`"
                else:
                    return f"{'e' if lowercase else 'E'}lement {index+1} of `{node.name}`"
            else:
                raise ValueError(f"pretty_name: unexpected node type {type(node)}")
        else:
            if isinstance(node, ir.Var):
                return f"var `{node.name}`"
            elif isinstance(node, ir.Field):
                name = node.name
                if isinstance(context, ir.Lookup):
                    name = f"{context.relation.name}.{name}"
                    prefix = "access"
                    lowercase = True
                elif isinstance(context, ir.Aggregate):
                    name = f"{context.aggregation.name}.{name}"
                    prefix = "access"
                    lowercase = True
                elif isinstance(context, ir.Rank):
                    name = "rank"
                    prefix = "access"
                    lowercase = True
                elif isinstance(context, ir.Relation):
                    name = f"{context.name}.{name}"
                    prefix = "field"
                elif not context:
                    prefix = "field"
                else:
                    raise ValueError(f"pretty_name: unexpected context type {type(context)}")

                if index < 0:
                    return f"{prefix} `{name}`"
                else:
                    return f"element {index+1} of {prefix} `{name}`"
            else:
                raise ValueError(f"pretty_name: unexpected node type {type(node)}")

@dataclass
class TypeError:
    """A type inference error."""
    msg: str
    node: ir.Node

    def __hash__(self):
        return hash((self.msg, self.node.id))

    def __eq__(self, other):
        return self.msg == other.msg and self.node.id == other.node.id

Typeable = PyUnion[ir.Value, ir.Type, ir.Var, ir.Field, TypeVar]

@dataclass
class TypeEnv:
    """Environment for type inference that tracks type bounds for each node."""

    # The model being type-checked.
    model: ir.Model

    # Diagnostics. For now, this is just strings.
    diags: OrderedSet[TypeError]

    # Maps node ids (and tuple index, as needed) to the type variables for their types.
    type_vars: Dict[Tuple[int, int, int], TypeVar]

    # How verbose to be with debug info, 0 for off.
    verbosity: int

    # Should we perform stricter checks on the inferred types.
    strict: bool

    def __init__(self, model: ir.Model, strict: bool, verbosity: int=0):
        super().__init__()
        self.model = model
        self.diags = OrderedSet[TypeError]()
        self.type_vars = {}
        self.strict = strict
        self.verbosity = verbosity

    def _complain(self, node: ir.Node, msg: str):
        """Report an error."""
        self.diags.add(TypeError(msg, node))

    def get_type_var(self, node: PyUnion[ir.Field, ir.Var], index: int = -1, context: PyUnion[ir.Lookup, ir.Aggregate, ir.Rank, None] = None) -> TypeVar:
        key = (node.id, index, context.id if context else node.id)
        if key not in self.type_vars:
            self.type_vars[key] = TypeVar(node, index, context)
        tv = self.type_vars[key]
        # TODO: Why does this sometimes fail?
        # assert tv.node == node, f"TypeVar {tv}, {tv.node} does not match node {node}"
        assert tv.index == index
        return tv

    def find_type_var(self, node: PyUnion[ir.Field, ir.Var], index: int = -1, context: PyUnion[ir.Lookup, ir.Aggregate, ir.Rank, None] = None) -> TypeVar:
        tv = self.get_type_var(node, index, context)
        return tv.find()

    def _get_type(self, t: PyUnion[ir.Literal, ir.PyValue]) -> ir.Type:
        if isinstance(t, str):
            return types.String
        elif isinstance(t, bool):
            return types.Bool
        elif isinstance(t, int):
            return types.Int
        elif isinstance(t, float):
            return types.Float
        # Note that the check for datetime must be before date since datetime <: date
        elif isinstance(t, datetime):
            return types.DateTime
        elif isinstance(t, date):
            return types.Date
        elif isinstance(t, ir.Literal):
            return t.type
        else:
            raise ValueError(f"Unexpected value {t} of type {type(t)}")

    def add_bound(self, lower: Typeable, upper: Typeable) -> None:
        lower_t = self._type_or_typevar(lower)
        upper_t = self._type_or_typevar(upper)
        self._add_bound(lower_t, upper_t)

    def _add_bound(self, lower: PyUnion[ir.Type, TypeVar], upper: PyUnion[ir.Type, TypeVar]):
        if isinstance(upper, ir.UnionType):
            for t in upper.types:
                self.add_bound(lower, t)
        elif isinstance(lower, ir.UnionType):
            for t in lower.types:
                self.add_bound(t, upper)
        elif isinstance(lower, ir.Type) and isinstance(upper, ir.Type):
            if lower == upper:
                return
            elif types.is_subtype(lower, upper):
                return
            else:
                self._complain(lower, f"Type {ir.type_to_string(lower)} is not a subtype of {ir.type_to_string(upper)}")
        elif isinstance(lower, ir.Type) and isinstance(upper, TypeVar):
            if not types.is_null(lower):
                upper.find().lower_bound.add(lower)
        elif isinstance(lower, TypeVar) and isinstance(upper, ir.Type):
            if not types.is_any(upper):
                lower.find().upper_bound.add(upper)
        else:
            assert isinstance(lower, TypeVar) and isinstance(upper, TypeVar), f"Unexpected types {type(lower)} and {type(upper)}"
            lower.find().node_super_types.add(upper.find())
            upper.find().node_sub_types.add(lower.find())

    def add_equality(self, t1: Typeable, t2: Typeable) -> None:
        tt1 = self._type_or_typevar(t1)
        tt2 = self._type_or_typevar(t2)
        self._add_equality(tt1, tt2)

    def _type_or_typevar(self, t: Typeable) -> PyUnion[ir.Type, TypeVar]:
        if t is None:
            return types.Null
        elif isinstance(t, (ir.Literal, str, bool, int, float, date, datetime)):
            return self._get_type(t)
        elif isinstance(t, (ir.Var, ir.Field)):
            return self.get_type_var(t)
        elif isinstance(t, tuple):
            return types.AnyList
        else:
            assert isinstance(t, ir.Type) or isinstance(t, TypeVar), f"Unexpected type {type(t)}"
            return t

    def _add_equality(self, t1: PyUnion[ir.Type, TypeVar], t2: PyUnion[ir.Type, TypeVar]):
        if isinstance(t1, ir.Type) and isinstance(t2, ir.Type):
            self.add_bound(t1, t2)
            self.add_bound(t2, t1)
        elif isinstance(t1, ir.Type) and isinstance(t2, TypeVar):
            # Add t1 as both a lower and upper bound of t2.
            if not types.is_any(t1):
                t2.find().upper_bound.add(t1)
            if not types.is_null(t1):
                t2.find().lower_bound.add(t1)
        elif isinstance(t1, TypeVar) and isinstance(t2, ir.Type):
            self._add_equality(t2, t1)
        else:
            assert isinstance(t1, TypeVar) and isinstance(t2, TypeVar)
            tt1 = t1.find()
            tt2 = t2.find()

            if tt1 != tt2:
                # Check that two type variables have compatible bounds.
                for lb in tt1.lower_bound:
                    for ub in tt2.upper_bound:
                        # TODO: why is that Not added?
                        if not types.is_subtype(lb, ub) and not types.is_subtype(ub, lb):
                            self._complain(t1.node, f"The types of {t1} and {t2} are not compatible: {ir.type_to_string(lb)} <: {t1}, but {t2} <: {ir.type_to_string(ub)}")
                for lb in tt2.lower_bound:
                    for ub in tt1.upper_bound:
                        # TODO: why is that Not added?
                        if not types.is_subtype(lb, ub) and not types.is_subtype(ub, lb):
                            self._complain(t2.node, f"The types of {t2} and {t1} are not compatible: {ir.type_to_string(lb)} <: {t2}, but {t1} <: {ir.type_to_string(ub)}")

                tt1.union(tt2)

    def dump(self, as_dot: bool=False):
        result = ""

        def _dump_title(title: str) -> str:
            if as_dot:
                return ""
            else:
               return f"\n{title}\n"

        def _dump_bound(lower: Typeable, upper: Typeable) -> str:
            if isinstance(lower, ir.Type):
                lower = ir.type_to_string(lower)
            if isinstance(upper, ir.Type):
                upper = ir.type_to_string(upper)

            if as_dot:
                return f"\t\"{lower}\" -> \"{upper}\"\n"
            else:
                return f"\t{lower} <: {upper}\n"

        def _dump_equality(t1: TypeVar, t2: TypeVar) -> str:
            if as_dot:
                return f"\t\"{t1}\" -> \"{t2}\" [dir=both, arrowtail=none, arrowhead=none, color=blue];\n"
            else:
                return f"\t{t1} == {t2}\n"

        # Match with the canonical representations
        result += _dump_title("Type vars:")
        for v in self._tyvars():
            if not v.is_canonical():
                result += _dump_equality(v, v.canonical_repr())

        # Match with the upper bound type vars
        result += _dump_title("Upper bound type vars:")
        for v in self._tyvars():
            if not v.is_canonical():
                continue
            for w in v.node_super_types:
                result += _dump_bound(v, w)

        # Match with the upper and lower type bounds
        result += _dump_title("Bounds for type vars:")
        for v in self._tyvars():
            if not v.is_canonical():
                continue
            if v.lower_bound:
                for typ in v.lower_bound:
                    result += _dump_bound(typ, v)
            if v.upper_bound:
                for typ in v.upper_bound:
                    result += _dump_bound(v, typ)

        if as_dot:
            result = "digraph G {\n" +\
                "\tnode [shape=box];\n" +\
                result +\
                "}"
        print(result)

    def _collapse_node_supertype_cycles(self):
        # Create a directed graph from the node_super_types relationships
        G = nx.DiGraph()

        for v in self._tyvars():
            v = v.find()
            for w in v.node_super_types:
                w = w.find()
                G.add_edge(v, w)

        # Equate all type vars in the same SCC.
        for scc in nx.strongly_connected_components(G):
            if len(scc) > 1:
                cycle = list(scc)
                for t in cycle[1:]:
                    cycle[0].union(t)

    def type_bounds_compatible(self, tv1: TypeVar, tv2: TypeVar) -> bool:
        # Check that two type variables have compatible bounds.
        for lb in tv1.lower_bound:
            for ub in tv2.upper_bound:
                if not types.is_subtype(lb, ub):
                    return False
        for lb in tv2.lower_bound:
            for ub in tv1.upper_bound:
                if not types.is_subtype(lb, ub):
                    return False
        return True

    def compute_type(self, tv: TypeVar, node: PyUnion[ir.Var, ir.Field], index: int, parent: Optional[ir.Node]) -> ir.Type:
        root = tv.find()

        t = root.compute_type()

        if types.is_null(t):
            upper = types.compute_lowest_bound(root.upper_bound)
            lower = types.compute_highest_bound(root.lower_bound)
            name = TypeVar.pretty_name(node, index, parent, lowercase=True)
            short_name = TypeVar.pretty_name(node, index, parent, lowercase=True, short=True)
            if lower and upper and not types.is_subtype(lower, upper):
                self._complain(node, f"Inferred an empty type for {name}: {ir.type_to_string(lower)} <: type({short_name}) <: {ir.type_to_string(upper)}.")
            else:
                self._complain(node, f"Inferred an empty type for {name}.")

        # We don't have any constraints that bound the type from above.
        # This means we should infer the type as Any, which is not useful.
        if self.strict and types.is_any(t):
            name = TypeVar.pretty_name(node, index, parent, lowercase=True)
            self._complain(node, f"Could not infer a type for {name} more specific than Any. This probably means ")

        return t

    def _propagate_bounds(self):
        """Propagate bounds along node_super_types edges until a fixpoint is reached."""

        worklist = ordered_set(*self._tyvars())
        while worklist:
            sub = worklist.pop()
            sub = sub.find()
            for sup in sub.node_super_types:
                sup = sup.find()
                if sup.upper_bound:
                    # Propagate upper bounds downward from supertype to subtype.
                    # We only need to propagate upper bounds downward if they are more precise.
                    # That is, if we have sub <: Number --> sup <: Int, we propagate the Int downward.
                    bound = []
                    for t in sup.upper_bound:
                        if any(types.is_proper_subtype(t, u) for u in sub.upper_bound):
                            bound.append(t)
                    if bound:
                        if self.verbosity > 2:
                            print(f"Unioning upper bound of {sub} with {sup} ub={ir.types_to_string(bound)}")
                            print(f" - old ub={ir.types_to_string(sub.upper_bound)}")
                            print(f" - new ub={ir.types_to_string(sub.upper_bound | bound)}")
                        n = len(sub.upper_bound)
                        sub.upper_bound |= bound
                        if n != len(sub.upper_bound):
                            worklist.add(sub)
                if sub.lower_bound:
                    # Propagate lower bounds upward from subtype to supertype.
                    # We only need to propagate lower bounds upward if they are more precise.
                    # That is, if we have Int :> sub --> Number :> sup, we propagate the Int upward.
                    bound = []
                    for t in sub.lower_bound:
                        if any(types.is_proper_subtype(u, t) for u in sub.lower_bound):
                            bound.append(t)
                    if bound:
                        if self.verbosity > 2:
                            print(f"Unioning lower bound of {sup} with {sub} lb={ir.types_to_string(bound)}")
                            print(f" - old lb={ir.types_to_string(sup.lower_bound)}")
                            print(f" - new lb={ir.types_to_string(sup.lower_bound | bound)}")
                        n = len(sup.lower_bound)
                        sup.lower_bound |= bound
                        if n != len(sup.lower_bound):
                            worklist.add(sup)

    def _tyvars(self) -> Iterable[TypeVar]:
        """Return all the type variables in the graph."""
        return self.type_vars.values()

    def _equivalent_tyvars(self, tv: TypeVar) -> Iterable[TypeVar]:
        """Return the set of type variables that are equivalent to the given type variable."""
        tv = tv.find()
        return [v for v in self._tyvars() if v.find() == tv]

    def _unify_instantiated_relations(self):
        v = UnifyInstantiatedRelations(self)
        self.model.accept(v)
        return v.changed

    def solve(self):
        """Solve the type constraints."""

        if self.verbosity:
            print("\n")
            if self.verbosity > 1:
                ir.dump(self.model)
            else:
                print(self.model)

            print("\n")
            print("Constraints before solving:")

            self.dump()

        # Collapse all the node_super_types cycles into equalities.
        self._collapse_node_supertype_cycles()
        # Propagate bounds.
        self._propagate_bounds()

        # Equate instantiated relations with their non-instantiated counterparts.
        while self._unify_instantiated_relations():
            # TODO This should be done incrementally on the new equivalent relations in the worklist.
            self._collapse_node_supertype_cycles()
            self._propagate_bounds()

        if self.verbosity:
            print("\n")
            print("Constraints after solving:")
            self.dump()

@dataclass
class UnifyInstantiatedRelations(visitor.Visitor):
    """
    Unify instantiated relations to their global counterparts.

    Consider the lookup R(a, b).

    Where R is defined as:
        R(x: Any, y: Any)
            overload R(x1: Int, y1: Int)
            overload R(x2: String, y2: String)

    We consider each variable and field to be a type variable.
    Fields without a context are "global" fields. Fields with a context are "instantiated" fields.
    We introduced type variables x' and y' for the instantiated fields of R at this lookup.
    The instantiated field type variables are created using the Lookup as the context;
    for instance, x' is env.get_type_var(x, context="lookup R(a, b)")

    CollectTypeConstraints should have added the following constraints:
        a = x'   # arguments of a lookup must match the instantiated fields
        b = y'
        x' <: x  # instantiated fields are subtypes of the global fields
        y' <: y
        x <: Any # declared type of the global fields
        y <: Any

    At this point, we would like to choose the overload that matches R.

    Suppose x'.compute_type() = Int.
    In this case, we search the overloads of R for one that matches and add the additional constraints:
        x' <: x1
        y' <: y1
    Adding these constraints will allow us to infer also that y'.compute_type() is also Int.

    Later, the ReconcileOverloads rewrite will replace the abstract relation R(x: Any, y: Any)
    used at the lookup with the concrete overload R(x1: Int, y1: Int).
    """
    env: TypeEnv = field(init=True)
    changed: bool = field(default=False)

    def computed_fields(self, relation: ir.Relation, context: PyUnion[ir.Lookup, ir.Aggregate, ir.Rank]) -> list[ir.Field]:
        computed_fields = []
        for f in relation.fields:
            if isinstance(f.type, ir.TupleType):
                fts = list(f.type.types)
                tvs = [self.env.get_type_var(f, i, context=context).find() for i in range(len(fts))]
                ts = [tv.compute_type() for tv in tvs]
                ts = [types.intersect(t, ft) for t, ft in zip(ts, fts)]
                t = ir.TupleType(tuple(ts))
            else:
                ft = f.type
                tv = self.env.get_type_var(f, context=context).find()
                t = tv.compute_type()
                t = types.intersect(t, ft)
            f2 = f.reconstruct(type=t)
            computed_fields.append(f2)
        return computed_fields

    def find_matching_relation(self, abstract_relation: ir.Relation, context: PyUnion[ir.Lookup, ir.Aggregate, ir.Rank]):
        """
        Find a relation with the same name and type signature that doesn't have a instantiated annotation.
        """
        # Find any overloads that might match the computed instantiated field types so far.
        # Now check if the computed types match.
        candidates = []
        for overload in abstract_relation.overloads:
            compatible = True
            for f1, f2 in zip(overload.fields, abstract_relation.fields):
                # This is complicated by the second-class handling of tuple types.
                if isinstance(f1.type, ir.TupleType) and isinstance(f2.type, ir.TupleType):
                    assert len(f1.type.types) == len(f2.type.types)
                    fts1 = list(f1.type.types)
                    fts2 = list(f2.type.types)
                    tvs1 = [self.env.get_type_var(f1, i).find() for i in range(len(fts1))]
                    tvs2 = [self.env.get_type_var(f2, i, context=context).find() for i in range(len(fts2))]
                else:
                    fts1 = [f1.type]
                    fts2 = [f2.type]
                    tvs1 = [self.env.get_type_var(f1).find()]
                    tvs2 = [self.env.get_type_var(f2, context=context).find()]

                for tv1, tv2, ft1, ft2 in zip(tvs1, tvs2, fts1, fts2):
                    if tv1 == tv2:
                        # Already equal.
                        continue
                    t1 = tv1.compute_type()
                    t2 = tv2.compute_type()
                    if not types.matches(t1, ft1):
                        if self.env.verbosity > 2:
                            print(f"Type mismatch: {ir.type_to_string(t1)} != {ir.type_to_string(ft1)}")
                        compatible = False
                        break
                    if not types.matches(t2, ft2):
                        if self.env.verbosity > 2:
                            print(f"Type mismatch: {ir.type_to_string(t2)} != {ir.type_to_string(ft2)}")
                        compatible = False
                        break
                    t1 = types.intersect(t1, ft1)
                    t2 = types.intersect(t2, ft2)
                    if not types.matches(t1, t2):
                        # Types don't match.
                        if self.env.verbosity > 2:
                            print(f"Types don't match: {ir.type_to_string(t1)} != {ir.type_to_string(t2)}")
                        compatible = False
                        break
                    else:
                        if self.env.verbosity > 2:
                            print(f"Types match: {tv1} ~ {tv2}")
                            print(f"Types match: {ir.type_to_string(t1)} ~ {ir.type_to_string(t2)}")
                    if not self.env.type_bounds_compatible(tv1, tv2):
                        # Bounds don't match.
                        if self.env.verbosity > 2:
                            print(f"Bounds don't match: {tv1} != {tv2}")
                        compatible = False
                        break
                if not compatible:
                    break

            if compatible:
                candidates.append(overload)
                if self.env.verbosity > 2:
                    print(f"Found candidate {overload} for {abstract_relation}")
            else:
                if self.env.verbosity > 2:
                    print(f"Not compatible: {overload} for {abstract_relation}")

        if len(candidates) > 1:
            # If there are multiple candidates, choose the most specific one.
            maximal_candidates = []
            for candidate in candidates:
                if not any(helpers.relation_is_proper_subtype(other, candidate) for other in candidates):
                    maximal_candidates.append(candidate)
            candidates = maximal_candidates

        if len(candidates) == 0:
            computed_types = [f.type for f in self.computed_fields(abstract_relation, context)]
            self.env._complain(abstract_relation, f"Relation `{abstract_relation.name}` (id={abstract_relation.id}) has no matching relation with types ({ir.types_to_string(computed_types)}).")
            return None
        elif len(candidates) == 1:
            # There is exactly one matching overload.
            # Equate the instantiated fields of the lookup with the global fields of the overload.
            overload = candidates[0]
            if self.env.verbosity > 2:
                print(f"Found matching relation {overload} for {abstract_relation}")
                print("\n\nENV before merging\n\n")
                self.env.dump()
                print("\n")
                print(f"Found matching relation {overload} for {abstract_relation}")
            for f1, f2 in zip(overload.fields, abstract_relation.fields):
                if isinstance(f1.type, ir.TupleType) and isinstance(f2.type, ir.TupleType):
                    assert len(f1.type.types) == len(f2.type.types)
                    tvs1 = [self.env.get_type_var(f1, i).find() for i in range(len(f1.type.types))]
                    tvs2 = [self.env.get_type_var(f2, i, context=context).find() for i in range(len(f2.type.types))]
                else:
                    tvs1 = [self.env.get_type_var(f1).find()]
                    tvs2 = [self.env.get_type_var(f2, context=context).find()]
                for tv1, tv2 in zip(tvs1, tvs2):
                    if tv1 != tv2:
                        self.env.add_equality(tv1, tv2)
                        self.changed = True
            if self.env.verbosity > 2:
                print("\n\nENV after merging\n\n")
                self.env.dump()
                print("\n")
        else:
            assert len(candidates) > 1
            # For now, if there are multiple candidates, just continue.
            # A later pass might narrow down the candidates.
            # self.env._complain(relation, f"Relation `{relation.name}` (id={relation.id}) has multiple matching relations.")
            pass

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]):
        if node.relation.overloads:
            self.find_matching_relation(node.relation, node)

        super().visit_lookup(node, parent)

    def visit_aggregate(self, node: ir.Aggregate, parent: Optional[ir.Node]):
        if node.aggregation.overloads:
            self.find_matching_relation(node.aggregation, node)
        super().visit_aggregate(node, parent)



import re

from relationalai.early_access.dsl.constants import VERBAL_PART_CONNECTION
from relationalai.early_access.dsl.core.utils import camel_to_snake, to_rai_way_string


# Re-adjusted code from dsl/utils.py - that version was used in the dsl
def build_relation_name_from_reading(reading):
    roles = re.findall(r'\{([^}]+)}', reading)
    text_without_concepts = re.sub(r'\{[^}]+}', '|', reading)
    verbal_parts = [text.strip() for text in text_without_concepts.split('|') if text.strip()]

    for rl, vp in zip(roles, verbal_parts):
        # Prefix
        if vp.endswith("-"):
            i = roles.index(rl)
            verbal_parts[i] = vp.replace("-", "-" + camel_to_snake(roles[i + 1]))
        # Postfix
        if vp.startswith("-"):
            i = roles.index(rl)
            postfix = vp.split(" ")[0]
            if verbal_parts[i] == postfix:
                verbal_parts.remove(vp)
            else:
                verbal_parts[i] = vp.replace(postfix, "")

    rel_name = ""
    if len(roles) == 1:
        if len(verbal_parts) > 0:
            rel_name = to_rai_way_string(verbal_parts[0])
        if not rel_name:
            rel_name = camel_to_snake(roles[0])
    elif len(roles) == 2:
        s_role = roles[1]
        if len(verbal_parts) > 0:
            rel_name = to_rai_way_string(verbal_parts[0])
            if verbal_parts[0].strip().startswith(("has", "is")):
                if not rel_name:
                    rel_name += camel_to_snake(s_role)
        else:
            rel_name = camel_to_snake(s_role)
    else:
        join_parts = [to_rai_way_string(verbal_parts[0], False)]
        for i in range(1, len(roles)):
            role = roles[i]
            if not join_parts[len(join_parts) - 1].endswith(camel_to_snake(role)):
                join_parts.append(to_rai_way_string(role.split(":")[0])) # take a role name instead of type if any
            if i < len(roles) - 1:
                join_parts.append(to_rai_way_string(verbal_parts[i]))
        rel_name = VERBAL_PART_CONNECTION.join(join_parts)
    return rel_name



"""
Type inference for the IR.
"""
from dataclasses import dataclass, field
from typing import Dict, Optional, Union as PyUnion, Tuple
import os

from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.metamodel import ir, types, visitor, compiler
from relationalai.early_access.metamodel.typer.type_env import TypeEnv, TypeVar

DEFAULT_VERBOSITY = int(os.getenv('PYREL_TYPER_VERBOSITY', '0'))
DEFAULT_REPORT_ERRORS = False if os.getenv('PYREL_TYPER_REPORT_ERRORS', '0') == '0' else True

# Type System
# -----------
#
# Scalar types are either value types or entity types, not both.
# Value types (transitively) extend the built-in scalar types for which
# `is_value_base_type` holds.
# All other scalar types are entity types.
#
# Subtyping of scalar is declared explicitly in the model, currently as the
# `super_types` field in `ScalarType`, but this may change.
#
# The type system supports unions and tuples of types.
# Tuple types are covariant in their element types.
# Union types support the usual subtyping relation (for instance, a member of a
# union is a subtype of the union).
# We could, but currently do not, introduce common supertypes for unions.
#
# The `Any` type is used as a placeholder to mean "infer this type". `Any` is a
# scalar type.
# In particular, `Any` is not implicitly a supertype of any other type.
#
# List types are not allowed in the model and should have been rewritten to
# tuple types by the `RewriteListTypes` pass.
#
# Set types in the model indicate the multiplicity of the given field. They are
# not strictly types.
# For inference purposes, the element type is used as the type of the field,
# however when substituting inference results back into the model, the inferred
# type is wrapped in a `SetType`.
# Set types must be sets of scalar type (including `Any`).
#
# Type inference works as follows:
# - For each application of a relation:
#   - for input fields, bound the argument type by the field type (that is, arg
#     <: field; this is the standard rule for argument passing, types flow from
#     the argument to the field)
#   - for non-input fields (which could be input or output), equate the argument
#     type with the corresponding field type in the relation (= is more
#     efficient than adding two bounds: arg <: field and field <: arg)
# - For each variable, bound its type with its declared type.
# - For each field, bound its type with its declared type.
# - For default values, bound the variable type with the type of the default value.
# - These constraints build a graph of equivalence classes of nodes (fields and
# - vars). An edge a->b indicates that a is a subtype of b.
#   Each nodes has an associated set of upper and lower bounding scalar types.
# - Propagate upper bounds through the graph from supertype to subtype.
# - Propagate lower bounds through the graph from suptype to supertype.
# - Collapse cycles (SCCs) in the graph into an equivalence class.
# - For each equivalence class, union the upper bounds of all types in the
#   class. This is the inferred upper bound. Check that the inferred upper bound
#   of a node is a supertype of the all lower bounds of the node.
# - TODO: Update to match the new overloading strategy
#   If the type of a field cannot be inferred to any type (as happens when the
#   field is not used in any application), equate the relation with other
#   relations of the same name and arity, if any. Then recompute the bound
#   again. This has the effect of treating relations of the same name and arity
#   as overloads of each other, but only when necessary.
# - Replace the types in the model with the intersection of the inferred upper
#   bound and the declared type. This strictly lowers types to be more precise.


# TODO: Add inference for value types for the builtins according to the "lattice algorithm".
# https://docs.google.com/document/d/1G62zNQhUxgi_vlLEUgB3uKBKTsoIG770Xx2jtGAshIk/edit?tab=t.0#heading=h.u82mjk2nuyvr
# That is, if we find a builtin relation (just arithmetic? and aggregates?) that uses some value type,
# add constraints to ensure the result of the operation is also a value type.
# Currently, this is broken because overloads are only defined for the value base types (Int, Float, etc.)
# We thus only infer these base types for the result of the operation.

# TODO: special case for constants when used with value types.
# Cost + 1 should be a Cost, not an Int.

# TODO: conditional constraints
# These should simplify the value type problem:
# Consider PosInt <: Int.
# Add constraints for each builtin (WLOG, +):
#   a + b = c
#   if a <: PosInt and b <: PosInt then c <: PosInt
# This should also simplify overloading, since we can just add conditional constraints for each overload.
#   if x' <: Int then x' <: x1 and y' <: y1  (where x1 and y1 are field type variables of the overload for Int)
#
# The environment needs to keep track of the conditional constraints.
# When ever we update a type variable bound, we check the conditional constraints.
#

class BuildSubst(visitor.DAGVisitor):
    """
    A visitor that computes the types of the nodes in the model.
    """
    env: TypeEnv
    strict: bool
    subst: Dict[TypeVar, OrderedSet[ir.Type]]

    def __init__(self, env: TypeEnv, strict: bool):
        super().__init__()
        self.env = env
        self.strict = strict
        self.subst = {}

    def compute_type(self, x: TypeVar, declared_type: ir.Type, node: PyUnion[ir.Var, ir.Field], parent: Optional[ir.Node]) -> ir.Type:
        t = self.env.compute_type(x, node, x.index, parent)
        name = TypeVar.pretty_name(node, x.index, parent)
        lower_name = TypeVar.pretty_name(node, x.index, parent, lowercase=True)
        if t is not None and not types.is_null(t):
            if isinstance(t, ir.UnionType):
                ts = []
                for t2 in t.types:
                    t3 = BuildSubst._wrap_type(declared_type, t2)
                    if types.is_subtype(t3, declared_type):
                        ts.append(t3)
                if len(ts) == 1:
                    new_type = ts[0]
                elif ts:
                    new_type = types.union(*ts)
                else:
                    new_type = BuildSubst._wrap_type(declared_type, t)
            elif not isinstance(t, ir.ScalarType):
                # TODO: what should be done here? what are the cases?
                self.env._complain(node, f"{name} is inferred to be a non-scalar type {t}. Variables must have scalar type.")
                new_type = BuildSubst._wrap_type(declared_type, t)
            else:
                new_type = BuildSubst._wrap_type(declared_type, t)
            if not types.is_subtype(new_type, declared_type):
                self.env._complain(node, f"{name} inferred to be type {ir.type_to_string(new_type)} which is not a subtype of the declared type {ir.type_to_string(declared_type)}.")
            return new_type
        else:
            if self.strict:
                self.env._complain(node, f"Could not infer a type for {lower_name}")
            return declared_type

    def visit_var(self, node: ir.Var, parent: Optional[ir.Node]=None) -> None:
        # TODO: what if the parent is none what should be done here?
        x = self.env.get_type_var(node)
        t = self.compute_type(x, node.type, node, parent)
        x = x.find()
        if x not in self.subst:
            self.subst[x] = ordered_set(t)
        else:
            self.subst[x].add(t)

    def visit_field(self, node: ir.Field, parent: Optional[ir.Node]=None) -> None:
        # Substitute the intersection of the inferred type with the declared type.
        if isinstance(node.type, ir.TupleType):
            new_types = []
            for i in range(len(node.type.types)):
                x = self.env.get_type_var(node, i)
                t = self.compute_type(x, node.type.types[i], node, parent)
                x = x.find()
                if x not in self.subst:
                    self.subst[x] = ordered_set(t)
                else:
                    self.subst[x].add(t)
                new_types.append(t)
            new_type = ir.TupleType(tuple(new_types))
        else:
            x = self.env.get_type_var(node)
            t = self.compute_type(x, node.type, node, parent)
            new_type = t
            x = x.find()
            if x not in self.subst:
                self.subst[x] = ordered_set(new_type)
            else:
                self.subst[x].add(new_type)

    # Intersect the declared type with the inferred type, special casing set types.
    @staticmethod
    def _wrap_type(declared: ir.Type, inferred: ir.Type) -> ir.Type:
        return inferred

@dataclass
class CollectTypeConstraints(visitor.DAGVisitor):
    """
    A visitor that collects type constraints on a model.
    """
    def __init__(self, env: TypeEnv):
        super().__init__()
        self.env = env

    def visit_scalartype(self, node: ir.ScalarType, parent: Optional[ir.Node]=None):
        # Do not recurse.
        pass

    def visit_listtype(self, node: ir.ListType, parent: Optional[ir.Node]=None):
        # Do not recurse.
        pass

    def visit_uniontype(self, node: ir.UnionType, parent: Optional[ir.Node]=None):
        # Do not recurse.
        pass

    def visit_field(self, node: ir.Field, parent: Optional[ir.Node]=None):
        # Do not constrain field types to be <: Any.
        if isinstance(node.type, ir.TupleType):
            for i in range(len(node.type.types)):
                x = self.env.get_type_var(node, i)
                if not types.is_any(node.type.types[i]):
                    self.env.add_bound(x, node.type.types[i])
        elif isinstance(node.type, ir.UnionType):
            for t in node.type.types:
                if not types.is_any(t):
                    self.env.add_bound(node, t)
        elif isinstance(node.type, ir.ScalarType):
            if not types.is_any(node.type):
                self.env.add_bound(node, node.type)
        # Do not recurse. No need to visit the type.
        pass

    def visit_var(self, node: ir.Var, parent: Optional[ir.Node]=None):
        # Do not constrain field types to be <: Any.
        if isinstance(node.type, ir.UnionType):
            for t in node.type.types:
                if not types.is_any(t):
                    self.env.add_bound(node, t)
        elif isinstance(node.type, ir.ScalarType):
            if not types.is_any(node.type):
                self.env.add_bound(node, node.type)
            if not types.is_abstract_type(node.type):
                self.env.add_equality(node, node.type)

        # Do not recurse. No need to visit the type.
        pass

    def visit_default(self, node: ir.Default, parent: Optional[ir.Node]=None):
        # The variable's type should be a supertype of the default value.
        self.env.add_bound(node.value, node.var)
        # Recurse to add the constraints on the variable.
        return super().visit_default(node, parent)

    def visit_literal(self, node: ir.Literal, parent: Optional[ir.Node]=None):
        # Do not recurse. No need to visit the type.
        pass

    def visit_loop(self, node: ir.Loop, parent: Optional[ir.Node]=None):
        # The iterator should be a number.
        self.env.add_bound(node.iter, types.Number)
        return super().visit_loop(node, parent)

    def visit_update(self, node: ir.Update, parent: Optional[ir.Node]=None):
        if len(node.args) == len(node.relation.fields):
            # We have update R(x,y,z) where R is declared (t,u,v)
            # Bound each arg by the declared type of the field.
            for f, arg in zip(node.relation.fields, node.args):
                if isinstance(f.type, ir.TupleType) and isinstance(arg, Tuple):
                    # TODO: add this to Checker
                    assert len(arg) != len(f.type.types)
                    for i in range(len(f.type.types)):
                        f_global = self.env.get_type_var(f, index=i)
                        if f.input:
                            # Flow from argument to input field.
                            self.env.add_bound(arg[i], f_global)
                        else:
                            # Flow from argument to field, and back.
                            self.env.add_equality(arg[i], f_global)
                else:
                    f_global = self.env.get_type_var(f)
                    if f.input:
                        # Flow from argument to input field.
                        self.env.add_bound(arg, f_global)
                    else:
                        # Flow from argument to field, and back.
                        self.env.add_bound(arg, f_global)

        return super().visit_update(node, parent)

    def visit_annotation(self, node: ir.Annotation, parent: Optional[ir.Node]=None):
        # Do not recurse. No need to visit the relation again.
        pass

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]=None):
        # The constraints here are in two parts:
        # 1. The relation used in this particular application must be a subtype
        #    of the global relation. This is handled by using separate type
        #    variables for the fields of the relation, one for the global
        #    relation (with no context) and one for the instantiation of the
        #    relation here (with the Lookup as context).
        #    If the relation is not overloaded, we can just equate the instantiation
        #    with the global relation.
        # 2. The arguments to the lookup must have the same types as the
        #    instantiation of the relation here.

        # 1. The instantiated fields are a subtype of the global fields.
        for f in node.relation.fields:
            if isinstance(f.type, ir.TupleType):
                for i in range(len(f.type.types)):
                    f_instantiated = self.env.get_type_var(f, index=i, context=node)
                    f_global = self.env.get_type_var(f, index=i, context=None)
                    if node.relation.overloads:
                        self.env.add_bound(f_instantiated, f_global)
                    else:
                        self.env.add_equality(f_instantiated, f_global)
            else:
                f_instantiated = self.env.get_type_var(f, context=node)
                f_global = self.env.get_type_var(f)
                if node.relation.overloads:
                    self.env.add_bound(f_instantiated, f_global)
                else:
                    self.env.add_equality(f_instantiated, f_global)

        # 2. The argument types are equal to the instantiated fields.
        if len(node.args) == len(node.relation.fields):
            for f, arg in zip(node.relation.fields, node.args):
                if isinstance(f.type, ir.TupleType) and isinstance(arg, Tuple):
                    if len(arg) == len(f.type.types):
                        for i in range(len(f.type.types)):
                            f_instantiated = self.env.get_type_var(f, index=i, context=node)
                            if f.input:
                                self.env.add_bound(arg[i], f_instantiated)
                            else:
                                self.env.add_equality(arg[i], f_instantiated)
                    else:
                        self.env._complain(node, f"The arity of the lookup is not equal to the arity of the relation: {len(arg)} != {len(f.type.types)}")
                else:
                    f_instantiated = self.env.get_type_var(f, context=node)
                    if f.input:
                        self.env.add_bound(arg, f_instantiated)
                    else:
                        self.env.add_equality(arg, f_instantiated)

            # Also, equality should bind both arguments to each other
            if node.relation.name == "=":
                self.env.add_equality(node.args[0], node.args[1])

        return super().visit_lookup(node, parent)

    def visit_aggregate(self, node: ir.Aggregate, parent: Optional[ir.Node]=None):
        # This is the same as the Lookup case above, but the structure of the aggregation
        # makes it a bit more complex.

        agg = node.aggregation

        if len(agg.fields) < 1:
            return

        inputs = []
        outputs = []
        for f in agg.fields:
            if f.input:
                inputs.append(f)
            else:
                outputs.append(f)

        # 1. The instantiated fields are a subtype of the global fields.
        for f in agg.fields:
            if isinstance(f.type, ir.TupleType):
                for i in range(len(f.type.types)):
                    f_instantiated = self.env.get_type_var(f, index=i, context=node)
                    f_global = self.env.get_type_var(f, index=i, context=None)
                    if agg.overloads:
                        self.env.add_bound(f_instantiated, f_global)
                    else:
                        self.env.add_equality(f_instantiated, f_global)
            else:
                f_instantiated = self.env.get_type_var(f, context=node)
                f_global = self.env.get_type_var(f)
                if agg.overloads:
                    self.env.add_bound(f_instantiated, f_global)
                else:
                    self.env.add_equality(f_instantiated, f_global)

        # Now let's wire up the types of the arguments with the instantiated fields.

        # Inputs and outputs.
        if len(node.args) == len(inputs) + len(outputs):
            for arg, f in zip(node.args, inputs + outputs):
                if isinstance(f.type, ir.TupleType) and isinstance(arg, Tuple):
                    # TODO: add this to Checker
                    assert len(arg) != len(f.type.types)
                    for i in range(len(f.type.types)):
                        f_instantiated = self.env.get_type_var(f, index=i, context=node)
                        if f.input:
                            self.env.add_bound(arg[i], f_instantiated)
                        else:
                            self.env.add_equality(arg[i], f_instantiated)
                else:
                    f_instantiated = self.env.get_type_var(f, context=node)
                    if f.input:
                        self.env.add_bound(arg, f_instantiated)
                    else:
                        self.env.add_equality(arg, f_instantiated)

        return super().visit_aggregate(node, parent)

    def visit_rank(self, node: ir.Rank, parent: Optional[ir.Node]=None):
        self.env.add_equality(node.result, types.Int)
        return super().visit_rank(node, parent)

@dataclass
class ReconcileOverloads(visitor.Rewriter):
    """
    Rewrite instantiated relations to their non-instantiated counterparts.
    This relies on UnifyInstantiatedRelations to find the non-instantiated relation.
    """
    env: TypeEnv = field(init=True)
    model: ir.Model = field(init=True)

    def find_matching_relation(self, abstract_relation: ir.Relation) -> Optional[ir.Relation]:
        """
        Find a non-instantiated relation with the same name and type signature.
        """
        # Find all non-instantiated relations in the model with the same name and arity that have
        # been unified with the given relation.
        # Unification was done by the UnifyInstantiatedRelations pass.
        candidates = []
        for r in abstract_relation.overloads:
            # If all the type variables match, consider it.
            if all(self.env.get_type_var(f1).find() == self.env.get_type_var(f2).find() for f1, f2 in zip(r.fields, abstract_relation.fields)):
                candidates.append(r)

        if len(candidates) == 0:
            self.env._complain(abstract_relation, f"Relation `{abstract_relation.name}` (id={abstract_relation.id}) has no matching overload.")
            return None
        elif len(candidates) == 1:
            return candidates[0]
        else:
            # If there's more than one matching candidate, union the fields types together.
            # TODO: or just filter out the candidates?
            # return abstract_relation.reconstruct(overloads=ordered_set(*candidates).frozen())
            new_fields = []
            for j, fields in enumerate(zip(*(r.fields for r in candidates))):
                ftypes = []
                name = abstract_relation.fields[j].name
                orig_type = abstract_relation.fields[j].type
                input = abstract_relation.fields[j].input
                for f in fields:
                    t = f.type
                    if isinstance(t, ir.ScalarType):
                        tv = self.env.get_type_var(f).find()
                        ftype = tv.compute_type()
                        ftype = types.intersect(ftype, t)
                        ftypes.append(ftype)
                    elif isinstance(t, ir.UnionType):
                        tv = self.env.get_type_var(f).find()
                        ftype = tv.compute_type()
                        for t2 in t.types:
                            ftype = types.intersect(ftype, t2)
                            ftypes.append(ftype)
                    elif isinstance(t, ir.TupleType):
                        # TODO: handle unions of tuples
                        for i, t2 in enumerate(t.types):
                            tv = self.env.find_type_var(f, i)
                            ftype = tv.compute_type()
                            ftype = types.intersect(ftype, t2)
                            ftypes.append(ftype)
                    else:
                        raise ValueError(f"Unexpected field type: {t}")
                new_type = types.union(orig_type, types.union(*ftypes))
                new_fields.append(ir.Field(name, new_type, input))

            new_relation = ir.Relation(
                name=abstract_relation.name,
                fields=tuple(new_fields),
                requires=abstract_relation.requires,
                annotations=abstract_relation.annotations,
                overloads=ordered_set().frozen(),
            )
            return new_relation

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node):
        if node.relation.overloads:
            matching_relation = self.find_matching_relation(node.relation)
            if matching_relation is not None:
                return node.reconstruct(relation=matching_relation)
        return node

    def handle_aggregate(self, node: ir.Aggregate, parent: ir.Node):
        if node.aggregation.overloads:
            matching_relation = self.find_matching_relation(node.aggregation)
            if matching_relation is not None:
                return node.reconstruct(aggregation=matching_relation)
        return node


@dataclass
class SubstituteTypes(visitor.Rewriter):
    """
    A visitor that substitutes types back into the model.
    """
    env: TypeEnv = field(init=True)
    subst: Dict[TypeVar, ir.Type] = field(init=True)

    def handle_var(self, node: ir.Var, parent: ir.Node) -> ir.Var:
        x = self.env.find_type_var(node)
        new_type = self.subst[x]
        disjuncts = []
        if isinstance(new_type, ir.UnionType):
            disjuncts = new_type.types
        else:
            disjuncts = [new_type]
        for t in disjuncts:
            if not isinstance(t, ir.ScalarType):
                self.env._complain(node, f"Variable {node.name} inferred to be a non-scalar type {ir.type_to_string(t)}. Variables must have scalar type.")
        return node.reconstruct(type=new_type)

    def handle_field(self, node: ir.Field, parent: ir.Node) -> ir.Field:
        # Substitute the intersection of the inferred type with the declared type.
        if isinstance(node.type, ir.TupleType):
            new_types = []
            for i in range(len(node.type.types)):
                x = self.env.find_type_var(node, i)
                t = self.subst[x]
                new_types.append(t)
            new_type = ir.TupleType(tuple(new_types))
        else:
            x = self.env.find_type_var(node)
            new_type = self.subst[x]
        return node.reconstruct(type=new_type)

@dataclass
class Typer(compiler.Pass):
    """
    A pass that performs type inference on a model.
    The pass also checks that the model is well-formed.
    Diagnostics are reported for ill-formed or ill-typed models.

    The main idea is to traverse the model and collect type constraints.
    These are then solved and substituted back into the model.
    """

    # Should we perform stricter checks on the inferred types?
    strict: bool = field(default=False, init=False)

    # How verbose to be with debug output, 0 is off.
    verbosity: int = field(default=DEFAULT_VERBOSITY, init=False)

    # Temporarily in in case there are bugs. Set this to default to False
    # if type inference is causing too many issues.
    report_errors: bool = field(default=DEFAULT_REPORT_ERRORS, init=False)

    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        if self.verbosity:
            print("\n")
            print("\nInitial model:")
            if self.verbosity > 1:
                ir.dump(model)
            else:
                print(model)

        # Type inference.
        env = TypeEnv(model, self.strict, self.verbosity)
        collector = CollectTypeConstraints(env)
        model.accept(collector)

        env.solve()

        # Substitute the types back into the model.
        build_subst = BuildSubst(env, self.strict)
        model.accept(build_subst)

        # Check that the substitutions are consistent.
        # Otherwise, we can get into a situation where we infer {S;T} and
        # substitute T in one place, but S in another.
        subst: Dict[TypeVar, ir.Type] = {}
        for x, ts in build_subst.subst.items():
            assert x == x.find()
            t = ts.list[0]
            for u in ts.list[1:]:
                if not types.is_subtype(t, u):
                    t = u
                elif not types.is_subtype(u, t):
                    env._complain(x.node, f"Type variable {x} is inferred to be {ir.type_to_string(t)} but it is used as the non-matching type {ir.type_to_string(u)}.")
            subst[x] = t

        do_subst = SubstituteTypes(env, subst)
        model2 = do_subst.walk(model)

        # Assert that there are no type errors
        if env.diags:
            error_count = len(env.diags)
            error_header = "TYPE ERROR\n" if error_count == 1 else f"{error_count} TYPE ERRORS\n"
            formatted_errors = [error_header] + [f"* (Node id={env.diags[i].node.id}) {env.diags[i].msg}" for i in range(error_count)]
            if self.report_errors:
                raise Exception("\n".join(formatted_errors))
            else:
                print("\n".join(formatted_errors))

        if self.verbosity:
            print("After substitution:")
            if self.verbosity > 1:
                ir.dump(model2)
            else:
                print(model2)

        reconcile = ReconcileOverloads(env, model2)
        model3 = reconcile.walk(model2)

        if self.verbosity:
            print("After reconcilation:")
            if self.verbosity > 1:
                ir.dump(model3)
            else:
                print(model3)

        return model3



from __future__ import annotations
from collections import defaultdict
import atexit
import re

from pandas import DataFrame
from typing import Any, Optional
import relationalai as rai

from relationalai import debugging
from relationalai.clients import result_helpers
from relationalai.early_access.metamodel import ir, factory as f, executor as e
from relationalai.early_access.lqp.compiler import Compiler
from lqp.proto.v1.transactions_pb2 import Transaction
from lqp import print as lqp_print, ir as lqp_ir
from relationalai.early_access.lqp.ir import convert_transaction, lqp_program_to_transaction
from relationalai.clients.config import Config
from relationalai.tools.constants import USE_DIRECT_ACCESS


class LQPExecutor(e.Executor):
    """Executes LQP using the RAI client."""

    def __init__(
        self,
        database: str,
        dry_run: bool = False,
        keep_model: bool = True,
        config: Config | None = None,
    ) -> None:
        super().__init__()
        self.database = database
        self.dry_run = dry_run
        self.keep_model = keep_model
        self.compiler = Compiler()
        self.config = config or Config()
        self._resources = None
        self._last_model = None
        self._last_sources_version = (-1, None)

    @property
    def resources(self):
        if not self._resources:
            with debugging.span("create_session"):
                self.dry_run |= bool(self.config.get("compiler.dry_run", False))
                resource_class = rai.clients.snowflake.Resources
                if self.config.get("use_direct_access", USE_DIRECT_ACCESS):
                    resource_class = rai.clients.snowflake.DirectAccessResources
                self._resources = resource_class(dry_run=self.dry_run, config=self.config)
                if not self.dry_run:
                    self.engine = self._resources.get_default_engine_name()
                    if not self.keep_model:
                        atexit.register(self._resources.delete_graph, self.database, True)
        return self._resources

    def check_graph_index(self):
        # Has to happen first, so self.dry_run is populated.
        resources = self.resources

        if self.dry_run:
            return

        from relationalai.early_access.builder.snowflake import Table
        table_sources = Table._used_sources
        if not table_sources.has_changed(self._last_sources_version):
            return

        model = self.database
        app_name = resources.get_app_name()
        engine_name = self.engine

        program_span_id = debugging.get_program_span_id()
        sources = [t._fqn for t in Table._used_sources]
        self._last_sources_version = Table._used_sources.version()

        assert self.engine is not None

        with debugging.span("poll_use_index", sources=sources, model=model, engine=engine_name):
            resources.poll_use_index(app_name, sources, model, self.engine, program_span_id)

    def report_errors(self, problems: list[dict[str, Any]], abort_on_error=True):
        from relationalai import errors
        all_errors = []
        undefineds = []
        pyrel_errors = defaultdict(list)
        pyrel_warnings = defaultdict(list)

        for problem in problems:
            message = problem.get("message", "")
            report = problem.get("report", "")
            # TODO: we need to build source maps
            # path = problem.get("path", "")
            # source_task = self._install_batch.line_to_task(path, problem["start_line"]) or task
            # source = debugging.get_source(source_task) or debugging.SourceInfo()
            source = debugging.SourceInfo()
            severity = problem.get("severity", "warning")
            code = problem.get("code")

            if severity in ["error", "exception"]:
                if code == "UNDEFINED_IDENTIFIER":
                    match = re.search(r'`(.+?)` is undefined', message)
                    if match:
                        undefineds.append((match.group(1), source))
                    else:
                        all_errors.append(errors.RelQueryError(problem, source))
                elif "overflowed" in report:
                    all_errors.append(errors.NumericOverflow(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_errors[problem["props"]["pyrel_id"]].append(problem)
                elif abort_on_error:
                    all_errors.append(errors.RelQueryError(problem, source))
            else:
                if code == "ARITY_MISMATCH":
                    errors.ArityMismatch(problem, source)
                elif code == "IC_VIOLATION":
                    all_errors.append(errors.IntegrityConstraintViolation(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_warnings[problem["props"]["pyrel_id"]].append(problem)
                else:
                    errors.RelQueryWarning(problem, source)

        if abort_on_error and len(undefineds):
            all_errors.append(errors.UninitializedPropertyException(undefineds))

        if abort_on_error:
            for pyrel_id, pyrel_problems in pyrel_errors.items():
                all_errors.append(errors.ModelError(pyrel_problems))

        for pyrel_id, pyrel_problems in pyrel_warnings.items():
            errors.ModelWarning(pyrel_problems)


        if len(all_errors) == 1:
            raise all_errors[0]
        elif len(all_errors) > 1:
            raise errors.RAIExceptionSet(all_errors)

    def execute_transaction(self, transaction: Transaction) -> DataFrame:
        if self.dry_run:
            return DataFrame()

        self.check_graph_index()

        raw_code = transaction.SerializeToString()

        # TODO have to run readonly for now
        raw_results = self.resources.exec_lqp(self.database, self.engine, raw_code, readonly=True, nowait_durable=True)
        df, errs = result_helpers.format_results(raw_results, None)  # Pass None for task parameter
        self.report_errors(errs)

        return df

    def execute(self, model: ir.Model, task:ir.Task, export_to:Optional[str]=None) -> DataFrame:
        self.check_graph_index()

        # TODO
        if export_to is not None:
            raise NotImplementedError("Export to is not supported in the LQP executor")

        model_txn = None
        if self._last_model != model:
            with debugging.span("compile", metamodel=model) as install_span:
                lqp_program = self.compiler.compile(model)
                model_txn = lqp_program_to_transaction(lqp_program, b"model")
                install_span["compile_type"] = "model"
                install_span["lqp"] = lqp_print.to_string(model_txn, lqp_print.ugly_config)
                self._last_model = model

        with debugging.span("compile", metamodel=task) as compile_span:
            query = f.compute_model(f.logical([task]))
            lqp_program = self.compiler.compile(query)
            query_txn = lqp_program_to_transaction(lqp_program, b"query")
            compile_span["compile_type"] = "query"
            compile_span["lqp"] = lqp_print.to_string(query_txn, lqp_print.ugly_config)

        txn = query_txn
        if model_txn is not None:
            # Merge the two LQP transactions into one. Ideally we would end up with all the
            # model bits in the persistent writes, and all the query bits in the local
            # writes. But for now we just use two separate epochs.
            model_epoch = model_txn.epochs[0]
            query_epoch = query_txn.epochs[0]
            txn = lqp_ir.Transaction(epochs=[model_epoch, query_epoch], meta=None)

        txn_proto = convert_transaction(txn)

        if self.dry_run:
            return DataFrame()

        return self.execute_transaction(txn_proto)
